from typing import Iterable, Tuple

import matplotlib.pyplot as plt
import statsmodels.api as sm
from statsmodels.discrete.conditional_models import ConditionalLogit, ConditionalResultsWrapper
from statsmodels.base.elastic_net import RegularizedResultsWrapper
from .base_fitters import ExpansionBasedFitter
from scipy.optimize import minimize
from scipy.special import logit, expit
import numpy as np
import pandas as pd
import psutil
from lifelines.fitters.coxph_fitter import CoxPHFitter
from pandarallel import pandarallel
from typing import Optional, List, Union
from matplotlib import colors as mcolors
from joblib import Parallel, delayed
import warnings
warnings.filterwarnings("ignore")

from .examples_utils.generate_simulations_data import generate_quick_start_df
from .utils import assert_fit

COLORS = list(mcolors.TABLEAU_COLORS.keys())
WORKERS = psutil.cpu_count(logical=False)


# Example:
#     ```py linenums="1"
#         from pydts.fitters import DataExpansionFitter
#         fitter = DataExpansionFitter()
#         fitter.fit(df=train_df, event_type_col='J', duration_col='X')
#         fitter.print_summary()
#     ```
#
# References:
#     [1] Lee, Minjung and Feuer, Eric J. and Fine, Jason P., "On the analysis of discrete time competing risks data", Biometrics (2018) doi: 10.1111/biom.12881

class DataExpansionFitter(ExpansionBasedFitter):
    """
    This class implements the estimation procedure of Lee et al. (2018) [1]. See also the Example section.
    """

    def __init__(self):
        super().__init__()
        self.models_kwargs = dict(family=sm.families.Binomial())

    def _fit_event(self, model_fit_kwargs={}):
        """
        This method fits a model for a GLM model for a specific event.

        Args:
            model_fit_kwargs (dict, Optional): Keyword arguments to pass to model.fit() method.

        Returns:
            fitted GLM model
        """
        model = sm.GLM.from_formula(formula=self.formula, data=self.expanded_df, **self.models_kwargs)
        return model.fit(**model_fit_kwargs)

    def fit(self,
            df: pd.DataFrame,
            event_type_col: str = 'J',
            duration_col: str = 'X',
            pid_col: str = 'pid',
            skip_expansion: bool = False,
            covariates: Optional[list] = None,
            formula: Optional[str] = None,
            models_kwargs: Optional[dict] = None,
            model_fit_kwargs: Optional[dict] = {}) -> dict:
        """
        This method fits a model to the discrete data.

        Args:
            df (pd.DataFrame): training data for fitting the model
            event_type_col (str): The event type column name (must be a column in df), Right censored sample (i) is indicated by event value 0, df.loc[i, event_type_col] = 0.
            duration_col (str): Last follow up time column name (must be a column in df).
            pid_col (str): Sample ID column name (must be a column in df).
            skip_expansion (boolean): Skips the dataframe expansion step. Use this option only if the provided dataframe (df) is already correctly expanded. When set to True, the df is expected to be in the format produced by the pydts.utils.get_expanded_df() method, as if it were applied to the unexpanded data.
            covariates (list, Optional): A list of covariates, all must be columns in df. Defaults to all the columns of df except event_type_col, duration_col, and pid_col.
            formula (str, Optional): Model formula to be fitted. Patsy format string.
            models_kwargs (dict, Optional): Keyword arguments to pass to model instance initiation.
            model_fit_kwargs (dict, Optional): Keyword arguments to pass to model.fit() method.

        Returns:
            event_models (dict): Fitted models dictionary. Keys - event names, Values - fitted models for the event.
        """

        if models_kwargs is not None:
            self.models_kwargs = models_kwargs

        if 'C' in df.columns:
            raise ValueError('C is an invalid column name, to avoid errors with categorical symbol C() in formula')
        self._validate_cols(df, event_type_col, duration_col, pid_col)
        if covariates is not None:
            cov_not_in_df = [cov for cov in covariates if cov not in df.columns]
            if len(cov_not_in_df) > 0:
                raise ValueError(f"Error during fit - missing covariates from df: {cov_not_in_df}")

        self.events = [c for c in sorted(df[event_type_col].unique()) if c != 0]
        self.covariates = [col for col in df if col not in [event_type_col, duration_col, pid_col]] \
                          if covariates is None else covariates
        self.times = sorted(df[duration_col].unique())

        if not skip_expansion:
            self.expanded_df = self._expand_data(df=df, event_type_col=event_type_col, duration_col=duration_col,
                                                 pid_col=pid_col)
        else:
            print('Skipping data expansion step, only use this option if the provided dataframe (df) is already correctly expanded.')
            self.expanded_df = df

        for event in self.events:
            cov = ' + '.join(self.covariates)
            _formula = f'j_{event} ~ {formula}' if formula is not None else \
                f'j_{event} ~ {cov} + C({duration_col}) -1 '
            self.formula = _formula
            self.event_models[event] = self._fit_event(model_fit_kwargs=model_fit_kwargs)
        return self.event_models

    def print_summary(self,
                      summary_func: str = "summary",
                      summary_kwargs: dict = {}) -> None:
        """
        This method prints the summary of the fitted models for all the events.

        Args:
            summary_func (str, Optional): print summary method of the fitted model type ("summary", "print_summary").
            summary_kwargs (dict, Optional): Keyword arguments to pass to the model summary function.

        Returns:
            None
        """
        for event, model in self.event_models.items():
            _summary_func = getattr(model, summary_func, None)
            if _summary_func is not None:
                print(f'\n\nModel summary for event: {event}')
                print(_summary_func(**summary_kwargs))
            else:
                print(f'Not {summary_func} function in event {event} model')

    def predict_hazard_jt(self,
                          df: pd.DataFrame,
                          event: Union[str, int],
                          t: Union[Iterable, int],
                          n_jobs: int = -1) -> pd.DataFrame:
        """
        This method calculates the hazard for the given event at the given time values if they were included in the training set of the event.

        Args:
            df (pd.DataFrame): samples to predict for
            event (Union[str, int]): event name
            t (np.array): times to calculate the hazard for
            n_jobs: number of CPUs to use, defualt to every available CPU

        Returns:
            df (pd.DataFrame): samples with the prediction columns
        """
        t = self._validate_t(t, return_iter=True)
        assert event in self.events, \
            f"Cannot predict for event {event} - it was not included during .fit()"
        self._validate_covariates_in_df(df.head())

        _t = np.array([t_i for t_i in t if (f'hazard_j{event}_t{t_i}' not in df.columns)])
        if len(_t) == 0:
            return df

        temp_df = df.copy()
        model = self.event_models[event]
        res = Parallel(n_jobs=n_jobs)(delayed(model.predict)(df[self.covariates].assign(X=c)) for c in t)
        temp_hazard_df = pd.concat(res, axis=1)
        temp_df[[f'hazard_j{event}_t{c_}' for c_ in t]] = temp_hazard_df.values
        return temp_df

    def get_beta_SE(self):
        """
        This function returns the Beta coefficients and their Standard Errors for all the events.

        Returns:
            se_df (pandas.DataFrame): Beta coefficients and Standard Errors Dataframe
        """

        full_table = pd.DataFrame()
        for event in self.events:
            summary = self.event_models[event].summary()
            summary_df = pd.DataFrame([x.split(',') for x in summary.tables[1].as_csv().split('\n')])
            summary_df.columns = summary_df.iloc[0]
            summary_df = summary_df.iloc[1:].set_index(summary_df.columns[0])
            summary_df.columns = pd.MultiIndex.from_product([[event], summary_df.columns])
            full_table = pd.concat([full_table, summary_df.iloc[-len(self.covariates):]], axis=1)
        return full_table

    def get_alpha_df(self):
        """
        This function returns the Alpha coefficients and their Standard Errors for all the events.

        Returns:
            se_df (pandas.DataFrame): Alpha coefficients and Standard Errors Dataframe
        """

        full_table = pd.DataFrame()
        for event in self.events:
            summary = self.event_models[event].summary()
            summary_df = pd.DataFrame([x.split(',') for x in summary.tables[1].as_csv().split('\n')])
            summary_df.columns = summary_df.iloc[0]
            summary_df = summary_df.iloc[1:].set_index(summary_df.columns[0])
            summary_df.columns = pd.MultiIndex.from_product([[event], summary_df.columns])
            full_table = pd.concat([full_table, summary_df.iloc[:-len(self.covariates)-1]], axis=1)
        return full_table



# Example:
#     ```py linenums="1"
#         from pydts.fitters import TwoStagesFitter
#         fitter = TwoStagesFitter()
#         fitter.fit(df=train_df, event_type_col='J', duration_col='X')
#         fitter.print_summary()
#     ```
#
# References:
#     [1] Meir, Tomer\*, Gutman, Rom\*, and Gorfine, Malka, "PyDTS: A Python Package for Discrete-Time Survival Analysis with Competing Risks" (2022)


class TwoStagesFitter(ExpansionBasedFitter):

    """
    This class implements the approach of Meir et al. (2022):
    """

    def __init__(self):
        super().__init__()
        self.alpha_df = pd.DataFrame()
        self.beta_models = {}
        self.beta_models_params_attr = 'params_'

    def _alpha_jt(self, x, df, y_t, beta_j, n_jt, t, event):
        # Alpha_jt optimization objective
        partial_df = df[df[self.duration_col] >= t]
        if isinstance(self.covariates, list):
            expit_add = np.dot(partial_df[self.covariates], beta_j)
        elif isinstance(self.covariates, dict):
            expit_add = np.dot(partial_df[self.covariates[event]], beta_j)
        else:
            raise ValueError
        return ((1 / y_t) * np.sum(expit(x + expit_add)) - (n_jt / y_t)) ** 2

    def _fit_event_beta(self, expanded_df, event, model=CoxPHFitter, model_kwargs={}, model_fit_kwargs={}):
        # Model fitting for conditional estimation of Beta_j for specific event
        if isinstance(self.covariates, list):
            strata_df = expanded_df[self.covariates + [f'j_{event}', self.duration_col]].copy()
        elif isinstance(self.covariates, dict):
            strata_df = expanded_df[self.covariates[event] + [f'j_{event}', self.duration_col]].copy()
        else:
            raise TypeError
        strata_df.loc[:, f'{self.duration_col}_copy'] = np.ones_like(expanded_df[self.duration_col])

        beta_j_model = model(**model_kwargs)
        if isinstance(self.covariates, list):
            beta_j_model.fit(df=strata_df[self.covariates + [f'{self.duration_col}', f'{self.duration_col}_copy', f'j_{event}']],
                             duration_col=f'{self.duration_col}_copy', event_col=f'j_{event}', strata=self.duration_col,
                             **model_fit_kwargs, batch_mode=False)
        elif isinstance(self.covariates, dict):
            beta_j_model.fit(df=strata_df[self.covariates[event] + [f'{self.duration_col}', f'{self.duration_col}_copy', f'j_{event}']],
                             duration_col=f'{self.duration_col}_copy', event_col=f'j_{event}', strata=self.duration_col,
                             **model_fit_kwargs, batch_mode=False)
        return beta_j_model

    def _fit_beta(self, expanded_df, events, model=CoxPHFitter, model_kwargs={}, model_fit_kwargs={}):
        # Model fitting for conditional estimation of Beta_j for all events
        _model_kwargs_per_event = np.any([event in model_kwargs.keys() for event in events])
        _model_fit_kwargs_per_event = np.any([event in model_fit_kwargs.keys() for event in events])
        beta_models = {}
        for event in events:
            _model_kwargs = model_kwargs[event] if _model_kwargs_per_event else model_kwargs
            _model_fit_kwargs = model_fit_kwargs[event] if _model_fit_kwargs_per_event else model_fit_kwargs
            beta_models[event] = self._fit_event_beta(expanded_df=expanded_df, event=event,
                                                      model=model, model_kwargs=_model_kwargs,
                                                      model_fit_kwargs=_model_fit_kwargs)
        return beta_models

    def fit(self,
            df: pd.DataFrame,
            covariates: Union[list, dict] = None,
            event_type_col: str = 'J',
            duration_col: str = 'X',
            pid_col: str = 'pid',
            skip_expansion: bool = False,
            x0: Union[np.array, int] = 0,
            fit_beta_kwargs: dict = {},
            verbose: int = 2,
            nb_workers: int = WORKERS) -> dict:
        """
        This method fits a model to the discrete data.

        Args:
            df (pd.DataFrame): training data for fitting the model
            covariates (list): list of covariates to be used in estimating the regression coefficients
            event_type_col (str): The event type column name (must be a column in df), Right-censored sample (i) is indicated by event value 0, df.loc[i, event_type_col] = 0.
            duration_col (str): Last follow up time column name (must be a column in df).
            pid_col (str): Sample ID column name (must be a column in df).
            skip_expansion (boolean): Skips the dataframe expansion step. Use this option only if the provided dataframe (df) is already correctly expanded. When set to True, the df is expected to be in the format produced by the pydts.utils.get_expanded_df() method, as if it were applied to the unexpanded data.
            x0 (Union[numpy.array, int], Optional): initial guess to pass to scipy.optimize.minimize function
            fit_beta_kwargs (dict, Optional): Keyword arguments to pass on to the estimation procedure.
            verbose (int, Optional): The verbosity level of pandaallel
            nb_workers (int, Optional): The number of workers to pandaallel. If not sepcified, defaults to the number of workers available.

        Returns:
            event_models (dict): Fitted models dictionary. Keys - event names, Values - fitted models for the event.
        """

        self._validate_cols(df, event_type_col, duration_col, pid_col)
        self.events = [c for c in sorted(df[event_type_col].unique()) if c != 0]
        if (covariates is not None):
            cov_not_in_df = []
            if isinstance(covariates, list):
                cov_not_in_df = [cov for cov in covariates if cov not in df.columns]
            elif isinstance(covariates, dict):
                for event in self.events:
                    event_cov_not_in_df = [cov for cov in covariates[event] if cov not in df.columns]
                    cov_not_in_df.extend(event_cov_not_in_df)
            if len(cov_not_in_df) > 0:
                raise ValueError(f"Error during fit - missing covariates from df: {cov_not_in_df}")

        #pandarallel.initialize(verbose=verbose, nb_workers=nb_workers)
        if covariates is None:
            covariates = [col for col in df if col not in [event_type_col, duration_col, pid_col]]
        self.covariates = covariates
        self.event_type_col = event_type_col
        self.duration_col = duration_col
        self.pid_col = pid_col
        self.times = sorted(df[duration_col].unique())

        if not skip_expansion:
            expanded_df = self._expand_data(df=df, event_type_col=event_type_col, duration_col=duration_col,
                                                 pid_col=pid_col)
        else:
            print('Skipping data expansion step, only use this option if the provided dataframe (df) is already correctly expanded.')
            expanded_df = df

        self.beta_models = self._fit_beta(expanded_df, self.events, **fit_beta_kwargs)

        y_t = (df[duration_col]
               .value_counts()
               .sort_index(ascending=False)  # each event count for its occurring time and the times before
               .cumsum()
               .sort_index()
               )
        n_jt = df.groupby([event_type_col, duration_col]).size().to_frame().reset_index()
        n_jt.columns = [event_type_col, duration_col, 'n_jt']

        for event in self.events:

            n_et = n_jt[n_jt[event_type_col] == event].copy()

            if isinstance(self.beta_models[event], CoxPHFitter):
                self.beta_models_params_attr = 'params_'
                _res = Parallel(n_jobs=nb_workers)(delayed(minimize)(self._alpha_jt, x0=x0,
                                                                     args=(df, y_t.loc[row[duration_col]],
                                                                           getattr(self.beta_models[event],
                                                                                   self.beta_models_params_attr),
                                                                           row['n_jt'],
                                                                           row[duration_col], event),
                                                                     method='BFGS',
                                                                     options={'gtol': 1e-7, 'eps': 1.5e-08,
                                                                              'maxiter': 200})
                                                                     for _, row in n_et.iterrows())
                n_et['success'] = Parallel(n_jobs=nb_workers)(delayed(lambda row: row.success)(val)
                                                              for val in _res)
                n_et['alpha_jt'] = Parallel(n_jobs=nb_workers)(delayed(lambda row: row.x[0])(val)
                                                               for val in _res)

            elif isinstance(self.beta_models[event], ConditionalResultsWrapper) or \
                    isinstance(self.beta_models[event], RegularizedResultsWrapper):
                self.beta_models_params_attr = 'params'
                for idx, row in n_et.iterrows():
                    _res = minimize(self._alpha_jt,
                                    x0=x0,
                                    args=(df,
                                          y_t.loc[row[duration_col]],
                                          getattr(self.beta_models[event], self.beta_models_params_attr),
                                          row['n_jt'],
                                          row[duration_col],
                                          event),
                                    method='BFGS',
                                    options={'gtol': 1e-7, 'eps': 1.5e-08, 'maxiter': 200})
                    n_et.loc[idx, 'success'] = _res.success
                    n_et.loc[idx, 'alpha_jt'] = _res.x[0]
            else:
                raise ValueError

            # n_et['opt_res'] = n_et.parallel_apply(lambda row: minimize(self._alpha_jt, x0=x0,
            #                         args=(df, y_t.loc[row[duration_col]], event_beta_params, row['n_jt'],
            #                         row[duration_col], event), method='BFGS',
            #                         options={'gtol': 1e-7, 'eps': 1.5e-08, 'maxiter': 200}), axis=1)
            # n_et['success'] = n_et['opt_res'].parallel_apply(lambda val: val.success)
            # n_et['alpha_jt'] = n_et['opt_res'].parallel_apply(lambda val: val.x[0])

            assert_fit(n_et, self.times[:-1], event_type_col=event_type_col, duration_col=duration_col)  # todo move basic input validation before any optimization
            self.event_models[event] = [self.beta_models[event], n_et]
            self.alpha_df = pd.concat([self.alpha_df, n_et], ignore_index=True)
        return self.event_models

    def print_summary(self,
                      summary_func: str = "print_summary",
                      summary_kwargs: dict = {}) -> None:
        """
        This method prints the summary of the fitted models for all the events.

        Args:
            summary_func (str, Optional): print summary method of the fitted model type ("summary", "print_summary").
            summary_kwargs (dict, Optional): Keyword arguments to pass to the model summary function.

        Returns:
            None
        """
        from IPython.display import display
        display(self.get_beta_SE())

        for event, model in self.event_models.items():
            print(f'\n\nModel summary for event: {event}')
            display(model[1].set_index([self.event_type_col, self.duration_col]))


    def plot_event_alpha(self, event: Union[str, int], ax: plt.Axes = None, scatter_kwargs: dict = {},
                         show=True, title=None, xlabel='t', ylabel=r'$\alpha_{jt}$', fontsize=18,
                         color: str = None, label: str = None, ticklabelsize: int = 15) -> plt.Axes:
        """
        This function plots a scatter plot of the $ alpha_{jt} $ coefficients of a specific event.

        Args:
            event (Union[str, int]): event name
            ax (matplotlib.pyplot.Axes, Optional): ax to use
            scatter_kwargs (dict, Optional): keywords to pass to the scatter function
            show (bool, Optional): if to use plt.show()
            title (str, Optional): axes title
            xlabel (str, Optional): axes xlabel
            ylabel (str, Optional): axes ylabel
            fontsize (int, Optional): axes title, xlabel, ylabel fontsize
            color (str, Optional): color name to use
            label (str, Optional): label name

        Returns:
            ax (matplotlib.pyplot.Axes): output figure
        """

        assert event in self.events, f"Cannot plot event {event} alpha - it was not included during .fit()"

        if ax is None:
            fig, ax = plt.subplots(1, 1)
        ax.tick_params(axis='both', which='major', labelsize=ticklabelsize)
        ax.tick_params(axis='both', which='minor', labelsize=ticklabelsize)
        title = r'$\alpha_{jt}$' + f' for event {event}' if title is None else title
        label = f'{event}' if label is None else label
        color = 'tab:blue' if color is None else color
        alpha_df = self.event_models[event][1]
        ax.scatter(alpha_df[self.duration_col].values, alpha_df['alpha_jt'].values, label=label,
                   color=color, **scatter_kwargs)
        ax.set_title(title, fontsize=fontsize)
        ax.set_xlabel(xlabel, fontsize=fontsize)
        ax.set_ylabel(ylabel, fontsize=fontsize)
        if show:
            plt.show()
        return ax

    def plot_all_events_alpha(self, ax: plt.Axes = None, scatter_kwargs: dict = {}, colors: list = COLORS,
                              show: bool = True, title: Union[str, None] = None, xlabel: str = 't',
                              ylabel: str = r'$\alpha_{jt}$', fontsize: int = 18, ticklabelsize: int = 15) -> plt.Axes:
        """
        This function plots a scatter plot of the $ alpha_{jt} $ coefficients of all the events.

        Args:
            ax (matplotlib.pyplot.Axes, Optional): ax to use
            scatter_kwargs (dict, Optional): keywords to pass to the scatter function
            colors (list, Optional): colors names
            show (bool, Optional): if to use plt.show()
            title (str, Optional): axes title
            xlabel (str, Optional): axes xlabel
            ylabel (str, Optional): axes ylabel
            fontsize (int, Optional): axes title, xlabel, ylabel fontsize

        Returns:
            ax (matplotlib.pyplot.Axes): output figure
        """
        if ax is None:
            fig, ax = plt.subplots(1, 1)
        ax.tick_params(axis='both', which='major', labelsize=ticklabelsize)
        ax.tick_params(axis='both', which='minor', labelsize=ticklabelsize)
        title = r'$\alpha_{jt}$' + f' for all events' if title is None else title
        for idx, (event, model) in enumerate(self.event_models.items()):
            label = f'{event}'
            color = colors[idx % len(colors)]
            self.plot_event_alpha(event=event, ax=ax, scatter_kwargs=scatter_kwargs, show=False, title=title,
                                  ylabel=ylabel, xlabel=xlabel, fontsize=fontsize, label=label, color=color,
                                  ticklabelsize=ticklabelsize)
        ax.legend()
        if show:
            plt.show()
        return ax

    def predict_hazard_jt(self,
                          df: pd.DataFrame,
                          event: Union[str, int],
                          t: Union[Iterable, int]) -> pd.DataFrame:
        """
        This method calculates the hazard for the given event at the given time values if they were included in the training set of the event.

        Args:
            df (pd.DataFrame): samples to predict for
            event (Union[str, int]): event name
            t (Union[Iterable, int]): times to calculate the hazard for

        Returns:
            df (pd.DataFrame): samples with the prediction columns
        """
        self._validate_covariates_in_df(df.head())
        t = self._validate_t(t, return_iter=True)
        assert event in self.events, \
            f"Cannot predict for event {event} - it was not included during .fit()"

        model = self.event_models[event]
        alpha_df = model[1].set_index(self.duration_col)['alpha_jt'].copy()

        _t = np.array([t_i for t_i in t if (f'hazard_j{event}_t{t_i}' not in df.columns)])
        if len(_t) == 0:
            return df
        temp_df = df.copy()
        if isinstance(self.covariates, list):
            beta_j_x = temp_df[self.covariates].dot(getattr(model[0], self.beta_models_params_attr))
        elif isinstance(self.covariates, dict):
            beta_j_x = temp_df[self.covariates[event]].dot(getattr(model[0], self.beta_models_params_attr))
        temp_df[[f'hazard_j{event}_t{c}' for c in _t]] = pd.concat(
            [self._hazard_inverse_transformation(alpha_df[c] + beta_j_x) for c in _t], axis=1).values
        return temp_df

    def _hazard_transformation(self, a: Union[int, np.array, pd.Series, pd.DataFrame]) -> \
            Union[int, np.array, pd.Series, pd.DataFrame]:
        """
        This function defines the transformation of the hazard function such that $ h ( \lambda_j (t | Z) ) = \alpha_{jt} + Z^{T} \beta_{j} $

        Args:
            a (Union[int, np.array, pd.Series, pd.DataFrame]):

        Returns:
            i (Union[int, np.array, pd.Series, pd.DataFrame]): the inverse function applied on a. $ h^{-1} (a)$
        """

        i = logit(a)
        return i

    def _hazard_inverse_transformation(self, a: Union[int, np.array, pd.Series, pd.DataFrame]) -> \
            Union[int, np.array, pd.Series, pd.DataFrame]:
        """
        This function defines the inverse transformation of the hazard function such that $\lambda_j (t | Z) = h^{-1} ( \alpha_{jt} + Z^{T} \beta_{j} )$

        Args:
            a (Union[int, np.array, pd.Series, pd.DataFrame]):

        Returns:
            i (Union[int, np.array, pd.Series, pd.DataFrame]): the inverse function applied on a. $ h^{-1} (a) $
        """
        i = expit(a)
        return i

    def get_beta_SE(self):
        """
        This function returns the Beta coefficients and their Standard Errors for all the events.

        Returns:
            se_df (pandas.DataFrame): Beta coefficients and Standard Errors Dataframe
        """
        se_df = pd.DataFrame()
        for event, model in self.beta_models.items():
            mdf = pd.concat([model.params_, model.standard_errors_], axis=1)
            mdf.columns = [f'j{event}_params', f'j{event}_SE']
            se_df = pd.concat([se_df, mdf], axis=1)
        return se_df

    def get_alpha_df(self):
        """
        This function returns the Alpha coefficients for all the events.

        Returns:
            alpha_df (pandas.DataFrame): Alpha coefficients Dataframe
        """

        alpha_df = pd.DataFrame()
        for event, model in self.event_models.items():
            model_alpha_df = model[1].set_index([self.event_type_col, self.duration_col])
            model_alpha_df.columns = pd.MultiIndex.from_product([[event], model_alpha_df.columns])
            alpha_df = pd.concat([alpha_df, model_alpha_df], axis=1)

        return alpha_df

    def plot_all_events_beta(self, ax: plt.Axes = None, colors: list = COLORS, show: bool = True,
                             title: Union[str, None] = None, xlabel: str = 'Value',  ylabel: str = r'$\beta_{j}$',
                             fontsize: int = 18, ticklabelsize: int = 15) -> plt.Axes:
        """
        This function plots the $ beta_{j} $ coefficients and standard errors of all the events.

        Args:
            ax (matplotlib.pyplot.Axes, Optional): ax to use
            colors (list, Optional): colors names
            show (bool, Optional): if to use plt.show()
            title (str, Optional): axes title
            xlabel (str, Optional): axes xlabel
            ylabel (str, Optional): axes ylabel
            fontsize (int, Optional): axes title, xlabel, ylabel fontsize
            ticklabelsize (int, Optional): axes xticklabels, yticklabels fontsize

        Returns:
            ax (matplotlib.pyplot.Axes): output figure
        """
        if ax is None:
            fig, ax = plt.subplots(1, 1)
        title = r'$\beta_{j}$' + f' for all events' if title is None else title
        ax.tick_params(axis='both', which='major', labelsize=ticklabelsize)
        ax.tick_params(axis='both', which='minor', labelsize=ticklabelsize)
        se_df = self.get_beta_SE()

        for idx, col in enumerate(se_df.columns):
            if idx % 2 == 1:
                continue
            y = np.arange((idx//2)*len(se_df), (1+(idx//2))*len(se_df))
            ax.errorbar(x=se_df.iloc[:, idx].values, y=y,
                       color=colors[idx % len(colors)], xerr=se_df.iloc[:, idx+1].values, label=f'{col}',
                       markersize=6, ls='', marker='o')

        yt = list(se_df.index) * (len(se_df.columns) // 2)
        ax.set_yticks(np.arange(0, len(yt)))
        ax.set_yticklabels(yt)
        ax.set_title(title, fontsize=fontsize)
        ax.set_xlabel(xlabel, fontsize=fontsize)
        ax.set_ylabel(ylabel, fontsize=fontsize)
        ax.grid()
        plt.gca().invert_yaxis()
        ax.legend()
        if show:
            plt.show()
        return ax


# def repetitive_fitters(rep: int, n_patients: int, n_cov: int, d_times: int, j_events: int, pid_col: str,
#                        drop_cols: Iterable = ("C", "T"),
#                        model1: ExpansionBasedFitter = DataExpansionFitter,
#                        model1_name="Lee",
#                        model2: ExpansionBasedFitter = TwoStagesFitter,
#                        model2_name: str = "Ours",
#                        allow_fails: int = 20,
#                        verbose: int = 2,
#                        real_coef_dict: dict = None,
#                        censoring_prob: float = 1.
#                        ) -> Tuple[dict, dict, pd.DataFrame]:
#     """
#     The function allows the user to generate N repetitions of model training (given data generating process),
#     to allow the user to compare the parameters stability and fitting time of the methods .
#
#     Args:
#         rep (int): number of repetitions to run the models
#         n_patients (int): number of sample to generate for each repetition
#         n_cov (int): number of covariates to generate for each repetition
#         d_times (int): how many times T to generate (i.e. times would be $t \in [1,\ldots,d]$)
#         j_events (int): number of events to generate for each repetition
#         pid_col (str): the name of the id column
#         test_size (float): the test size (percentage) for train/test splitting
#         drop_cols (Iterable):
#             Columns that shouldn't be visible for the model, from the generated sample.
#             default is drop the real event time (T) and censoring time (C)
#         model1 (ExpansionBasedFitter): Typically Lee et al. 2018 [1] model (DataExpansionFitter).
#             but can be any of base type ExpansionBasedFitter
#         model1_name (str): the name of the first model
#         model2 (ExpansionBasedFitter): Typically our suggested method (TwoStagesFitter).
#             but can be any of base type ExpansionBasedFitter
#         model2_name  (str): the name of the second model
#         allow_fails (int): number of allowed failed repetitions.
#             I.e. the method would run up to rep + allow_fails times.
#         verbose (int): verbosity level for the pandarallel module
#         real_coef_dict (dict): dictionary which represent the real coefficients to be generated for each repetition
#         censoring_prob (float): The probability to use the censoring method for each round of generation
#
#     Returns:
#         rep_dict (dict): Dictionary which contains for each round (key)
#             its beta/alpha/real coefficients dataframe (value).
#         times (dict): Dictionary which  contains for each model (key) a list of the training times (value).
#         ret_df (pd.DataFrame): Dataframe which contains for each time in d times, the average amount of events for it.
#
#     """
#
#     from .examples_utils.plots import compare_beta_models_for_example
#     from tqdm import trange
#     from time import time
#     assert real_coef_dict is not None, "The user should supply the coefficients of the experiment"
#     rep_dict = {}
#     times = {model1_name: [], model2_name: []}
#     counts_df_list = []
#     final = 0
#     failed = 0
#     for samp in trange(rep+allow_fails):
#         try:
#             patients_df = generate_quick_start_df(n_patients=n_patients, n_cov=n_cov, d_times=d_times,
#                                                   j_events=j_events,
#                                                   pid_col=pid_col, seed=samp, real_coef_dict=real_coef_dict,
#                                                   censoring_prob=censoring_prob )
#             counts_df = patients_df[patients_df['X'] <= d_times].groupby(['J', 'X']).size().to_frame(samp)
#             assert not (counts_df.reset_index()['X'].value_counts() < (j_events + 1)).any(), "Not enough events"
#             counts_df_list.append(counts_df)
#             drop_cols = pd.Index(drop_cols)
#             start_1 = time()
#             fitter = model1()
#             fitter.fit(df=patients_df.drop(drop_cols, axis=1))
#             end_1 = time()
#             start_2 = time()
#             new_fitter = model2()
#             if isinstance(new_fitter, TwoStagesFitter):
#                 new_fitter.fit(df=patients_df.drop(drop_cols, axis=1), verbose=verbose)
#             else:
#                 new_fitter.fit(df=patients_df.drop(drop_cols, axis=1))
#             end_2 = time()
#             times[model1_name].append(end_1 - start_1)
#             times[model2_name].append(end_2 - start_2)
#             res_dict = compare_beta_models_for_example(fitter.event_models,
#                                                        new_fitter.event_models, real_coef_dict=real_coef_dict)
#             rep_dict[samp] = res_dict
#             final += 1
#             if final == rep:
#                 break
#         except Exception as e:
#             print(e)
#             failed += 1
#             print(f'Failed to fit sample {samp+1}, fail #{failed}')
#             continue
#     print(f'final: {final}')
#     ret_df = pd.concat(counts_df_list, axis=1).fillna(0).mean(axis=1).apply(np.ceil).to_frame()
#     return rep_dict, times, ret_df


class TwoStagesFitterExact(TwoStagesFitter):

    def __init__(self):
        super().__init__()
        self.beta_models_params_attr = 'params'

    def _fit_event_beta(self, expanded_df, event, model=ConditionalLogit, model_kwargs={}, model_fit_kwargs={}):
        # Model fitting for conditional estimation of Beta_j for specific event
        if isinstance(self.covariates, dict):
            _covs = self.covariates[event]
        else:
            _covs = self.covariates

        beta_j_model = ConditionalLogit(endog=expanded_df[f'j_{event}'],
                                        exog=expanded_df[_covs],
                                        groups=expanded_df[self.duration_col],
                                        **model_kwargs)

        if ('alpha' in model_fit_kwargs.keys()):
            # Use 0 <= L1_wt <= 1 parameter to switch between L2 (L1_wt = 0) and L1 (L1_wt = 1) or elastic net.
            # alpha is the the penalty weight.
            beta_j_model = beta_j_model.fit_regularized(**model_fit_kwargs)
        else:
            beta_j_model = beta_j_model.fit(**model_fit_kwargs)

        return beta_j_model

    def get_beta_SE(self):
        """
        This function returns the Beta coefficients and their Standard Errors for all the events.

        Returns:
            se_df (pandas.DataFrame): Beta coefficients and Standard Errors Dataframe
        """

        full_table = pd.DataFrame()
        for event in self.events:
            if isinstance(self.beta_models[event], RegularizedResultsWrapper):
                _p = self.beta_models[event].params.copy()
                _p.name = 'coef'
                full_table = pd.concat([full_table,
                                        pd.concat([_p], keys=[event], axis=1)],
                                       axis=1)
            else:
                summary = self.beta_models[event].summary()
                summary_df = pd.DataFrame([x.split(',') for x in summary.tables[1].as_csv().split('\n')])
                summary_df.columns = summary_df.iloc[0]
                summary_df = summary_df.iloc[1:].set_index(summary_df.columns[0])
                summary_df.columns = pd.MultiIndex.from_product([[event], summary_df.columns])
                full_table = pd.concat([full_table, summary_df.iloc[-len(self.covariates):]], axis=1)
        return full_table



from .fitters import TwoStagesFitter, TwoStagesFitterExact
import pandas as pd
import numpy as np
from typing import Optional, List, Union
import psutil
from .utils import get_expanded_df
from joblib import Parallel, delayed


WORKERS = psutil.cpu_count(logical=False)


class MarginalTwoStagesFitter(TwoStagesFitter):

    def fit(self,
            expanded_df: pd.DataFrame,
            covariates: List = None,
            event_type_col: str = 'J',
            duration_col: str = 'X',
            pid_col: str = 'pid',
            x0: Union[np.array, int] = 0,
            fit_beta_kwargs: dict = {},
            verbose: int = 2,
            nb_workers: int = WORKERS) -> dict:
        """
        This method estimates only the parameters of the covariates (beta_j) without the time parameters (alpha_jt). Note that the expanded discrete-time data is expected as an input (see the Methods section of PyDTS documentation and pydts.utils.get_expanded_df).

        Args:
            expanded_df (pd.DataFrame): expanded training data for fitting the model
            covariates (list): list of covariates to be used in estimating the regression coefficients
            event_type_col (str): The event type column name (must be a column in df), Right-censored sample (i) is indicated by event value 0, df.loc[i, event_type_col] = 0.
            duration_col (str): Last follow up time column name (must be a column in df).
            pid_col (str): Sample ID column name (must be a column in df).
            x0 (Union[numpy.array, int], Optional): initial guess to pass to scipy.optimize.minimize function
            fit_beta_kwargs (dict, Optional): Keyword arguments to pass on to the estimation procedure.
            verbose (int, Optional): The verbosity level of pandaallel
            nb_workers (int, Optional): The number of workers to pandaallel. If not sepcified, defaults to the number of workers available.

        Returns:
            event_models (dict): Fitted models dictionary. Keys - event names, Values - fitted models for the event.
        """

        self._validate_cols(expanded_df, event_type_col, duration_col, pid_col)
        self.events = [c for c in sorted(expanded_df[event_type_col].unique()) if c != 0]

        if covariates is None:
            covariates = [col for col in expanded_df if col not in [event_type_col, duration_col, pid_col]]
        self.covariates = covariates
        self.event_type_col = event_type_col
        self.duration_col = duration_col
        self.pid_col = pid_col
        self.times = sorted(expanded_df[duration_col].unique())

        self.expanded_df = expanded_df

        self.beta_models = self._fit_beta(expanded_df, self.events, **fit_beta_kwargs)

        return self.beta_models


class MarginalTwoStagesFitterExact(TwoStagesFitterExact):

    def fit(self,
            expanded_df: pd.DataFrame,
            covariates: List = None,
            event_type_col: str = 'J',
            duration_col: str = 'X',
            pid_col: str = 'pid',
            x0: Union[np.array, int] = 0,
            fit_beta_kwargs: dict = {},
            verbose: int = 2,
            nb_workers: int = WORKERS) -> dict:
        """
        This method estimates only the parameters of the covariates (beta_j) without the time parameters (alpha_jt). Note that the expanded discrete-time data is expected as an input (see the Methods section of PyDTS documentation and pydts.utils.get_expanded_df).

        Args:
            expanded_df (pd.DataFrame): expanded training data for fitting the model
            covariates (list): list of covariates to be used in estimating the regression coefficients
            event_type_col (str): The event type column name (must be a column in df), Right-censored sample (i) is indicated by event value 0, df.loc[i, event_type_col] = 0.
            duration_col (str): Last follow up time column name (must be a column in df).
            pid_col (str): Sample ID column name (must be a column in df).
            x0 (Union[numpy.array, int], Optional): initial guess to pass to scipy.optimize.minimize function
            fit_beta_kwargs (dict, Optional): Keyword arguments to pass on to the estimation procedure.
            verbose (int, Optional): The verbosity level of pandaallel
            nb_workers (int, Optional): The number of workers to pandaallel. If not sepcified, defaults to the number of workers available.

        Returns:
            event_models (dict): Fitted models dictionary. Keys - event names, Values - fitted models for the event.
        """

        self._validate_cols(expanded_df, event_type_col, duration_col, pid_col)
        self.events = [c for c in sorted(expanded_df[event_type_col].unique()) if c != 0]

        if covariates is None:
            covariates = [col for col in expanded_df if col not in [event_type_col, duration_col, pid_col]]
        self.covariates = covariates
        self.event_type_col = event_type_col
        self.duration_col = duration_col
        self.pid_col = pid_col
        self.times = sorted(expanded_df[duration_col].unique())

        self.expanded_df = expanded_df

        self.beta_models = self._fit_beta(expanded_df, self.events, **fit_beta_kwargs)

        return self.beta_models


class BaseSISTwoStages(object):

    """
    This class implements the principled sure independence screening (PSIS) process of Zhao et al. (2012) for discrete-time data using the TwoStagesFitter and data-driven threshold.
    """

    def __init__(self):
        self.threshold = None
        self.marginal_estimates_df = pd.DataFrame()
        self.df = pd.DataFrame()
        self.expanded_df = pd.DataFrame()
        self.permuted_df = pd.DataFrame()
        self.permuted_expanded_df = pd.DataFrame()
        self.events = None
        self.covariates = None
        self.event_type_col = None
        self.duration_col = None
        self.pid_col = None
        self.times = None
        self.null_model_df = None
        self.final_model = None
        self.chosen_covariates_j = None
        self.chosen_covariates = None
        self.TwoStagesFitter_type = ""

    def fit_marginal_model(self,
                           expanded_df,
                           covariate: str,
                           event_type_col: str = 'J',
                           duration_col: str = 'X',
                           pid_col: str = 'pid',
                           x0: Union[np.array, int] = 0,
                           fit_beta_kwargs: dict = {},
                           verbose: int = 2,
                           nb_workers: int = 1):
        """
        This method fits a marginal model to data using a single covariate. Note that the expanded discrete-time data is expected as an input (see the Methods section of PyDTS documentation and pydts.utils.get_expanded_df).

        Args:
            expanded_df (pd.DataFrame): expanded training data for fitting the model
            covariate (str): a single covariate to be used in estimating the regression coefficients
            event_type_col (str): The event type column name (must be a column in df), Right-censored sample (i) is indicated by event value 0, df.loc[i, event_type_col] = 0.
            duration_col (str): Last follow up time column name (must be a column in df).
            pid_col (str): Sample ID column name (must be a column in df).
            x0 (Union[numpy.array, int], Optional): initial guess to pass to scipy.optimize.minimize function
            fit_beta_kwargs (dict, Optional): Keyword arguments to pass on to the estimation procedure.
            verbose (int, Optional): The verbosity level of pandaallel
            nb_workers (int, Optional): The number of workers to pandaallel. If not sepcified, defaults to the number of workers available.

        Returns:
            result (pd.DataFrame): Estimated parameter and standard errors. TwoStagesFitter.get_beta_SE() output.
        """

        if self.events is None:
            self.events = [c for c in sorted(expanded_df[event_type_col].unique()) if c != 0]

        if self.TwoStagesFitter_type == 'Exact':
            marginal_model = MarginalTwoStagesFitterExact()
        else:
            marginal_model = MarginalTwoStagesFitter()

        marginal_model.fit(
            expanded_df=expanded_df[[pid_col, covariate, event_type_col, duration_col, 'j_0'] +
                                    [f'j_{e}' for e in self.events]],
            covariates=[covariate],
            event_type_col=event_type_col,
            duration_col=duration_col,
            pid_col=pid_col,
            x0=x0,
            fit_beta_kwargs=fit_beta_kwargs,
            verbose=verbose,
            nb_workers=nb_workers)

        result = marginal_model.get_beta_SE()
        del marginal_model
        return result

    def get_marginal_estimates(self,
                               expanded_df,
                               covariates: Union[List, dict] = None,
                               event_type_col: str = 'J',
                               duration_col: str = 'X',
                               pid_col: str = 'pid',
                               verbose: int = 2,
                               x0: Union[np.array, int] = 0,
                               fit_beta_kwargs: dict = {},
                               nb_workers: int = WORKERS):

        """
        This method fits a marginal model to data to each of the covariates. Note that the expanded discrete-time data is expected as an input (see the Methods section of PyDTS documentation and pydts.utils.get_expanded_df).

        Args:
            expanded_df (pd.DataFrame): expanded training data for fitting the model
            covariates (list): list of covariates to estimate the marginal regression coefficient for.
            event_type_col (str): The event type column name (must be a column in df), Right-censored sample (i) is indicated by event value 0, df.loc[i, event_type_col] = 0.
            duration_col (str): Last follow up time column name (must be a column in df).
            pid_col (str): Sample ID column name (must be a column in df).
            verbose (int, Optional): The verbosity level of pandaallel
            x0 (Union[numpy.array, int], Optional): initial guess to pass to scipy.optimize.minimize function
            fit_beta_kwargs (dict, Optional): Keyword arguments to pass on to the estimation procedure.
            nb_workers (int, Optional): The number of workers to pandaallel. If not sepcified, defaults to the number of workers available.

        Returns:
            results_df (pd.DataFrame): Estimated parameters and standard errors of the marginal models. A concatenation of all the TwoStagesFitter.get_beta_SE() outputs.
        """

        if self.events is None:
            self.events = [c for c in sorted(expanded_df[event_type_col].unique()) if c != 0]

        if covariates is None:
            covariates = [col for col in expanded_df if col not in ([event_type_col, duration_col, pid_col, 'j_0'] +
                          [f'j_{e}' for e in self.events])]

        parallel = Parallel(n_jobs=nb_workers, verbose=verbose)
        results_df = pd.DataFrame()
        if isinstance(covariates, list):
            _results = parallel(delayed(self.fit_marginal_model)(expanded_df, cov,
                                                                 event_type_col, duration_col, pid_col,
                                                                 x0, fit_beta_kwargs, verbose, nb_workers)
                                                                 for cov in covariates)
            results_df = pd.concat(_results)
        elif isinstance(covariates, dict):
            raise ValueError("Please provide a list of covariates for the marginal testing, including the union of all possible options across all risks.")
        # elif isinstance(covariates, dict):
        #     for event in self.events:
        #         _results = parallel(delayed(self.fit_marginal_model)(expanded_df, cov,
        #                                                              event_type_col, duration_col, pid_col,
        #                                                              x0, fit_beta_kwargs, verbose, nb_workers)
        #                                                              for cov in covariates[event])
        #         event_results_df = pd.concat(_results)
        #         results_df = pd.concat([results_df, event_results_df], axis=1)

        return results_df.astype(float)

    def permute_df(self,
                   df,
                   event_type_col: str = 'J',
                   duration_col: str = 'X',
                   pid_col: str = 'pid',
                   seed: int = None):

        """
        This method applies random permutation on the event-time and event-type columns of the training data such that the covariates are decoupled from the outcome; the permuted data follow the null model.

        Args:
            df (pd.DataFrame): training data for fitting the model
            event_type_col (str): The event type column name (must be a column in df), Right-censored sample (i) is indicated by event value 0, df.loc[i, event_type_col] = 0.
            duration_col (str): Last follow up time column name (must be a column in df).
            pid_col (str): Sample ID column name (must be a column in df).
            seed (int, Optional): pseudo random state.

        Returns:
            permuted_df (pd.DataFrame): null model data.
        """

        permuted_df = df.copy()
        np.random.seed(seed)
        permuted_index = np.random.permutation(permuted_df.index)
        permuted_df.loc[:, duration_col] = df.loc[permuted_index, duration_col].values
        permuted_df.loc[:, event_type_col] = df.loc[permuted_index, event_type_col].values
        self.permuted_df = permuted_df
        self.permuted_expanded_df = get_expanded_df(df=self.permuted_df,
                                                    event_type_col=event_type_col,
                                                    duration_col=duration_col,
                                                    pid_col=pid_col)
        return permuted_df

    def get_data_driven_threshold(self,
                                  df,
                                  covariates: List = None,
                                  quantile: float = 1,
                                  event_type_col: str = 'J',
                                  duration_col: str = 'X',
                                  pid_col: str = 'pid',
                                  x0: Union[np.array, int] = 0,
                                  fit_beta_kwargs: dict = {},
                                  verbose: int = 2,
                                  nb_workers: int = WORKERS,
                                  seed: int = None):

        """
        This method calculates a data-driven threshold for each risk. It fits marginal models to the permuted data and returns the required quantile of the absolute values of the coefficients estimated from the null model.

        Args:
            df (pd.DataFrame): training data for fitting the model
            covariates (list): list of covariates to estimate the marginal regression coefficient for.
            quantile (float): represents the quantile of the absolute values of the coefficients from the null model that determines the data-driven threshold. Defaults to 1, which corresponds to the maximum absolute value of the null model's coefficients.
            event_type_col (str): The event type column name (must be a column in df), Right-censored sample (i) is indicated by event value 0, df.loc[i, event_type_col] = 0.
            duration_col (str): Last follow up time column name (must be a column in df).
            pid_col (str): Sample ID column name (must be a column in df).
            x0 (Union[numpy.array, int], Optional): initial guess to pass to scipy.optimize.minimize function
            fit_beta_kwargs (dict, Optional): Keyword arguments to pass on to the estimation procedure.
            verbose (int, Optional): The verbosity level of pandaallel
            nb_workers (int, Optional): The number of workers to pandaallel. If not sepcified, defaults to the number of workers available.
            seed (int): pseudo random state.

        Returns:
            threshold (pd.Series): Estimated thresholds.
        """

        if self.events is None:
            self.events = [c for c in sorted(df[event_type_col].unique()) if c != 0]
        if covariates is None:
            covariates = [col for col in df if col not in [event_type_col, duration_col, pid_col]]

        self.permute_df(df=df, event_type_col=event_type_col,
                        duration_col=duration_col, pid_col=pid_col, seed=seed)
        self.null_model_df = self.get_marginal_estimates(expanded_df=self.permuted_expanded_df,
                                                         covariates=covariates,
                                                         event_type_col=event_type_col,
                                                         duration_col=duration_col,
                                                         pid_col=pid_col,
                                                         verbose=verbose,
                                                         x0=x0,
                                                         fit_beta_kwargs=fit_beta_kwargs,
                                                         nb_workers=nb_workers)

        _params_cols = self._get_params_cols_from_res_df(self.null_model_df)
        self.threshold = np.quantile(self.null_model_df[_params_cols].abs().values, q=quantile)
        return self.threshold

    def fit(self,
            df: pd.DataFrame,
            threshold: float = None,
            quantile: float = 1,
            covariates: List = None,
            event_type_col: str = 'J',
            duration_col: str = 'X',
            pid_col: str = 'pid',
            x0: Union[np.array, int] = 0,
            fit_beta_kwargs: dict = {},
            verbose: int = 2,
            nb_workers: int = WORKERS,
            seed: int = None,
            fit_final_model: bool = True):

        """
        This method performs the principled sure independence screening (PSIS) process of Zhao et al. (2012) for discrete-time data with data-driven threshold.

        Args:
            df (pd.DataFrame): training data for fitting the model
            threshold (float): a user defined threshold. Defaults to None, i.e. data-driven threshold
            quantile (float): the quantile of the absolute values of the coefficients from the null model that determines the data-driven threshold. Only in use when threshold = None. Defaults to 1, which corresponds to the maximum absolute value of the null model's coefficients.
            covariates (list): list of covariates to estimate the marginal regression coefficient for.
            event_type_col (str): The event type column name (must be a column in df), Right-censored sample (i) is indicated by event value 0, df.loc[i, event_type_col] = 0.
            duration_col (str): Last follow up time column name (must be a column in df).
            pid_col (str): Sample ID column name (must be a column in df).
            x0 (Union[numpy.array, int], Optional): initial guess to pass to scipy.optimize.minimize function
            fit_beta_kwargs (dict, Optional): Keyword arguments to pass on to the estimation procedure.
            verbose (int, Optional): The verbosity level of pandaallel
            nb_workers (int, Optional): The number of workers to pandaallel. If not sepcified, defaults to the number of workers available.
            seed (int): pseudo random state.
            fit_final_model (boolean): True if to fit and return the TwoStagesFitter with the selected covariates.

        Returns:
            final_model (TwoStagesFitter): estimated model with the chosen covariates after PSIS.
        """

        self.events = [c for c in sorted(df[event_type_col].unique()) if c != 0]
        if covariates is None:
            covariates = [col for col in df if col not in [event_type_col, duration_col, pid_col]]
        self.covariates = covariates
        self.event_type_col = event_type_col
        self.duration_col = duration_col
        self.pid_col = pid_col
        self.times = sorted(df[duration_col].unique())

        if threshold is not None:
            self.threshold = threshold
        else:
            self.threshold = self.get_data_driven_threshold(df=df,
                                                            covariates=covariates,
                                                            quantile=quantile,
                                                            event_type_col=event_type_col,
                                                            duration_col=duration_col,
                                                            pid_col=pid_col,
                                                            x0=x0,
                                                            fit_beta_kwargs=fit_beta_kwargs,
                                                            verbose=verbose,
                                                            nb_workers=nb_workers,
                                                            seed=seed)
        self.df = df
        self.expanded_df = get_expanded_df(df=df,
                                           event_type_col=event_type_col,
                                           duration_col=duration_col,
                                           pid_col=pid_col)

        self.marginal_estimates_df = self.get_marginal_estimates(expanded_df=self.expanded_df,
                                                                 covariates=covariates,
                                                                 event_type_col=event_type_col,
                                                                 duration_col=duration_col,
                                                                 pid_col=pid_col,
                                                                 verbose=verbose,
                                                                 x0=x0,
                                                                 fit_beta_kwargs=fit_beta_kwargs,
                                                                 nb_workers=nb_workers)

        chosen_covariates = []
        chosen_covariates_j = {}

        _params_cols = self._get_params_cols_from_res_df(self.marginal_estimates_df)

        for c in _params_cols:
            if self.TwoStagesFitter_type == 'Exact':
                event = int(c[0])
            else:
                event = int(c[1:].split('_')[0])
            chosen_covariates_j[event] = self.marginal_estimates_df[self.marginal_estimates_df[c].abs() >= self.threshold].index.tolist()
            chosen_covariates.extend(chosen_covariates_j[event])

        self.chosen_covariates = sorted(np.unique(chosen_covariates))
        self.chosen_covariates_j = chosen_covariates_j

        if fit_final_model:
            if self.TwoStagesFitter_type == 'Exact':
                self.final_model = TwoStagesFitterExact()
            else:
                self.final_model = TwoStagesFitter()
            self.final_model.fit(df=df,
                                 covariates=self.chosen_covariates_j,
                                 event_type_col=event_type_col,
                                 duration_col=duration_col,
                                 pid_col=pid_col,
                                 x0=x0,
                                 fit_beta_kwargs=fit_beta_kwargs,
                                 verbose=verbose,
                                 nb_workers=nb_workers)

        return self.final_model

    def _get_params_cols_from_res_df(self, res_df):
        if self.TwoStagesFitter_type == 'Exact':
            _params_cols = [c for c in res_df.columns if '   coef   ' in c]
        else:
            _params_cols = [c for c in res_df.columns if 'params' in c]
        return _params_cols


class SISTwoStagesFitter(BaseSISTwoStages):

    def __init__(self):
        super().__init__()
        self.TwoStagesFitter_type = 'CoxPHFitter'


class SISTwoStagesFitterExact(BaseSISTwoStages):

    def __init__(self):
        super().__init__()
        self.TwoStagesFitter_type = 'Exact'


from typing import Iterable, Optional

import pandas as pd
import numpy as np
from scipy.special import expit


# def get_expanded_df(df, event_type_col='J', duration_col='X', pid_col='pid'):
#     """
#     This function gets a dataframe describing each sample the time of the observed events,
#     and returns an expanded dataframe as explained in [1]-[2].
#     Right censoring is allowed and must be marked as event type 0.
#
#     :param df: original dataframe (pd.DataFrame)
#     :param event_type_col: event type column name (str)
#     :param duration_col: time column name (str)
#     :param pid_col: patient id column name (str)
#
#     :return: result_df: expanded dataframe
#
#     References:
#         [1] Meir, Tomer and Gorfine, Malka, "Discrete-time Competing-Risks Regression with or without Penalization", https://arxiv.org/abs/2303.01186
#         [2] Meir, Tomer and Gutman, Rom and Gorfine, Malka "PyDTS: A Python Package for Discrete-Time Survival (Regularized) Regression with Competing Risks", https://arxiv.org/abs/2204.05731
#     """

def get_expanded_df(
        df: pd.DataFrame,
        event_type_col: str = 'J',
        duration_col: str = 'X',
        pid_col: str = 'pid') -> pd.DataFrame:
    """
    Expands a discrete-time survival dataset into a long-format dataframe suitable for modeling. This function receives a dataframe where each row corresponds to a subject with observed  event type and duration. It returns an expanded dataframe where each subject is represented  by multiple rows, one for each time point up to their observed time. Right censoring is allowed and should be indicated by event type 0.

    Args:
        df (pd.DataFrame): Original input dataframe containing one row per subject.
        event_type_col (str): Name of the column indicating event type. Censoring is marked by 0.
        duration_col (str): Name of the column indicating event or censoring time.
        pid_col (str): Name of the column indicating subject/patient ID.

    Returns:
        pd.DataFrame: Expanded dataframe in long format, with one row per subject-time pair.
    """
    unique_times = df[duration_col].sort_values().unique()
    result_df = df.reindex(df.index.repeat(df[duration_col]))
    result_df[duration_col] = result_df.groupby(pid_col).cumcount() + 1
    # drop times that didn't happen
    result_df.drop(index=result_df.loc[~result_df[duration_col].isin(unique_times)].index, inplace=True)
    result_df.reset_index(drop=True, inplace=True)
    last_idx = result_df.drop_duplicates(subset=[pid_col], keep='last').index
    events = sorted(df[event_type_col].unique())
    result_df.loc[last_idx, [f'j_{e}' for e in events]] = pd.get_dummies(
        result_df.loc[last_idx, event_type_col]).values
    result_df[[f'j_{e}' for e in events]] = result_df[[f'j_{e}' for e in events]].fillna(0)
    result_df[f'j_0'] = 1 - result_df[[f'j_{e}' for e in events if e > 0]].sum(axis=1)
    return result_df


def compare_models_coef_per_event(first_model: pd.Series,
                                  second_model: pd.Series,
                                  real_values: np.array,
                                  event: int,
                                  first_model_label:str = "first",
                                  second_model_label:str = "second"
                                  ) -> pd.DataFrame:
    event_suffix = f"_{event}"
    assert (first_model.index == second_model.index).all(), "All index should be the same"
    models = pd.concat([first_model.to_frame(first_model_label),
                        second_model.to_frame(second_model_label)], axis=1)
    models.index += event_suffix
    real_values_s = pd.Series(real_values, index=models.index)

    return pd.concat([models, real_values_s.to_frame("real")], axis=1)


def present_coefs(res_dict):
    from IPython.display import display
    for coef_type, events_dict in res_dict.items():
        print(f"for coef: {coef_type.capitalize()}")
        df = pd.concat([temp_df for temp_df in events_dict.values()])
        display(df)


def get_real_hazard(df, real_coef_dict, times, events):
    a_t = {event: {t: real_coef_dict['alpha'][event](t) for t in times} for event in events}
    b = pd.concat([df.dot(real_coef_dict['beta'][j]) for j in events], axis=1, keys=events)

    for j in events:
        df[[f'hazard_j{j}_t{t}' for t in times]] = pd.concat([expit(a_t[j][t] + b[j]) for t in times],
                                                             axis=1).values
    return df


def assert_fit(event_df, times, event_type_col='J', duration_col='X'):
    if not event_df['success'].all():
        problematic_times = event_df.loc[~event_df['success'], duration_col].tolist()
        event = event_df[event_type_col].max()  # all the events in the dataframe are the same
        raise RuntimeError(f"Number of observed events at some time points are too small. Consider collapsing neighbor time points."
                           f"\n See https://tomer1812.github.io/pydts/UsageExample-RegroupingData/ for more details.")
    if not np.all([t in event_df[duration_col].values for t in times]):
        event = event_df[event_type_col].max()  # all the events in the dataframe are the same
        problematic_times = pd.Index(event_df[duration_col]).symmetric_difference(times).tolist()
        raise RuntimeError(f"Number of observed events at some time points are too small. Consider collapsing neighbor time points."
                           f"\n See https://tomer1812.github.io/pydts/UsageExample-RegroupingData/ for more details.")


def create_df_for_cif_plots(df: pd.DataFrame, field: str,
                            covariates: Iterable,
                            vals: Optional[Iterable] = None,
                            quantiles: Optional[Iterable] = None,
                            zero_others: Optional[bool] = False
                            ) -> pd.DataFrame:
    """
    This method creates df for cif plot, where it zeros

    Args:
        df (pd.DataFrame): Dataframe which we yield the statiscal propetrics (means, quantiles, etc) and stacture
        field (str): The field which will represent the change
        covariates (Iterable): The covariates of the given model
        vals (Optional[Iterable]): The values to use for the field
        quantiles (Optional[Iterable]): The quantiles to use as values for the field
        zero_others (bool): Whether to zero the other covarites or to zero them

    Returns:
        df (pd.DataFrame): A dataframe that contains records per value for cif ploting
    """

    cov_not_fitted = [cov for cov in covariates if cov not in df.columns]
    assert len(cov_not_fitted) == 0, \
        f"Required covariates are missing from df: {cov_not_fitted}"

    df_for_ploting = df.copy()
    if vals is not None:
        pass
    elif quantiles is not None:
        vals = df_for_ploting[field].quantile(quantiles).values
    else:
        raise NotImplemented("Only Quantiles or specific values is supported")
    temp_series = []
    template_s = df_for_ploting.iloc[0][covariates].copy()
    if zero_others:
        impute_val = 0
    else:
        impute_val = df_for_ploting[covariates].mean().values
    for val in vals:
        temp_s = template_s.copy()
        temp_s[covariates] = impute_val
        temp_s[field] = val
        temp_series.append(temp_s)

    return pd.concat(temp_series, axis=1).T


import numpy as np
from scipy.special import expit
from typing import Union
import pandas as pd


class EventTimesSampler(object):

    def __init__(self, d_times: int, j_event_types: int):
        """
        This class implements sampling procedure for discrete event times and censoring times for given observations.

        Args:
            d_times (int): number of possible event times
            j_event_types (int): number of possible event types
        """

        self.d_times = d_times
        self.times = range(1, self.d_times + 2)  # d + 1 for administrative censoring
        self.j_event_types = j_event_types
        self.events = range(1, self.j_event_types + 1)

    def _validate_prob_dfs_list(self, dfs_list: list, numerical_error_tolerance: float = 0.001) -> list:
        for df in dfs_list:
            if (((df < (0-numerical_error_tolerance)).any().any()) or ((df > (1+numerical_error_tolerance)).any().any())):
                raise ValueError("The chosen sampling parameters result in invalid probabilities for event j at time t")
            # Only fixes numerical errors smaller than the tolerance size
            df.clip(0, 1, inplace=True)
        return dfs_list

    def calculate_hazards(self, observations_df: pd.DataFrame, hazard_coefs: dict, events: list = None) -> list:
        """
        Calculates the hazard function for the observations given the hazard coefficients.

        Args:
            observations_df (pd.DataFrame): Dataframe with observations covariates.
            coefficients (dict): time coefficients and covariates coefficients for each event type.

        Returns:
            hazards_dfs (list): A list of dataframes, one for each event type, with the hazard function at time t to each of the observations.
        """
        events = events if events is not None else self.events
        a_t = {}
        for event in events:
            if callable(hazard_coefs['alpha'][event]):
                a_t[event] = {t: hazard_coefs['alpha'][event](t) for t in range(1, self.d_times + 1)}
            else:
                a_t[event] = {t: hazard_coefs['alpha'][event][t-1] for t in range(1, self.d_times + 1)}
        b = pd.concat([observations_df.dot(hazard_coefs['beta'][j]) for j in events], axis=1, keys=events)
        hazards_dfs = [pd.concat([expit((a_t[j][t] + b[j]).astype(float)) for t in range(1, self.d_times + 1)],
                                  axis=1, keys=(range(1, self.d_times + 1))) for j in events]
        return hazards_dfs

    def calculate_overall_survival(self, hazards: list, numerical_error_tolerance: float = 0.001) -> pd.DataFrame:
        """
        Calculates the overall survival function given the hazards.

        Args:
            hazards (list): A list of hazards dataframes for each event type (as returned from EventTimesSampler.calculate_hazards function).
            numerical_error_tolerance (float): Tolerate numerical errors of probabilities up to this value.

        Returns:
            overall_survival (pd.Dataframe): The overall survival functions.
        """
        if (((sum(hazards)) > (1 + numerical_error_tolerance)).sum().sum() > 0):
            raise ValueError("The chosen sampling parameters result in negative values of the overall survival function")
        sum_hazards = sum(hazards).clip(0, 1)
        overall_survival = pd.concat([pd.Series(1, index=hazards[0].index),
                                      (1 - sum_hazards).cumprod(axis=1).iloc[:, :-1]], axis=1)
        overall_survival.columns += 1
        return overall_survival

    def calculate_prob_event_at_t(self, hazards: list, overall_survival: pd.DataFrame,
                                  numerical_error_tolerance: float = 0.001) -> list:
        """
        Calculates the probability for event j at time t.

        Args:
            hazards (list): A list of hazards dataframes for each event type (as returned from EventTimesSampler.calculate_hazards function)
            overall_survival (pd.Dataframe): The overall survival functions
            numerical_error_tolerance (float): Tolerate numerical errors of probabilities up to this value.

        Returns:
            prob_event_at_t (list): A list of dataframes, one for each event type, with the probability of event occurrance at time t to each of the observations.
        """
        prob_event_at_t = [hazard * overall_survival for hazard in hazards]
        prob_event_at_t = self._validate_prob_dfs_list(prob_event_at_t, numerical_error_tolerance)
        return prob_event_at_t

    def calculate_prob_event_j(self, prob_j_at_t: list, numerical_error_tolerance: float = 0.001) -> list:
        """
        Calculates the total probability for event j.

        Args:
            prob_j_at_t (list): A list of dataframes, one for each event type, with the probability of event occurrance at time t to each of the observations.
            numerical_error_tolerance (float): Tolerate numerical errors of probabilities up to this value.

        Returns:
            total_prob_j (list): A list of dataframes, one for each event type, with the total probability of event occurrance to each of the observations.
        """
        total_prob_j = [prob.sum(axis=1) for prob in prob_j_at_t]
        total_prob_j = self._validate_prob_dfs_list(total_prob_j, numerical_error_tolerance)
        return total_prob_j

    def calc_prob_t_given_j(self, prob_j_at_t, total_prob_j, numerical_error_tolerance=0.001):
        """
        Calculates the conditional probability for event occurrance at time t given J_i=j.

        Args:
            prob_j_at_t (list): A list of dataframes, one for each event type, with the probability of event occurrance at time t to each of the observations.
            total_prob_j (list): A list of dataframes, one for each event type, with the total probability of event occurrance to each of the observations.
            numerical_error_tolerance (float): Tolerate numerical errors of probabilities up to this value.

        Returns:
            conditional_prob (list): A list of dataframes, one for each event type, with the conditional probability of event occurrance at t given event type j to each of the observations.
        """
        conditional_prob = [prob.div(sumj, axis=0) for prob, sumj in zip(prob_j_at_t, total_prob_j)]
        conditional_prob = self._validate_prob_dfs_list(conditional_prob, numerical_error_tolerance)
        return conditional_prob

    def sample_event_times(self, observations_df: pd.DataFrame,
                                 hazard_coefs: dict,
                                 covariates: Union[list, None] = None,
                                 events: Union[list, None] = None,
                                 seed: Union[int, None] = None) -> pd.DataFrame:
        """
        Sample event type and event occurance times.

        Args:
            observations_df (pd.DataFrame): Dataframe with observations covariates.
            covariates (list): list of covariates name, must be a subset of observations_df.columns
            coefficients (dict): time coefficients and covariates coefficients for each event type.
            seed (int, None): numpy seed number for pseudo random sampling.

        Returns:
            observations_df (pd.DataFrame): Dataframe with additional columns for sampled event time (T) and event type (J).
        """
        np.random.seed(seed)
        if covariates is None:
            covariates = [c for c in observations_df.columns if c not in ['X', 'T', 'C', 'J']]
        events = events if events is not None else self.events
        cov_df = observations_df[covariates]
        hazards = self.calculate_hazards(cov_df, hazard_coefs, events=events)
        overall_survival = self.calculate_overall_survival(hazards)
        probs_j_at_t = self.calculate_prob_event_at_t(hazards, overall_survival)
        total_prob_j = self.calculate_prob_event_j(probs_j_at_t)
        probs_t_given_j = self.calc_prob_t_given_j(probs_j_at_t, total_prob_j)
        sampled_jt = self.sample_jt(total_prob_j, probs_t_given_j)
        if 'J' in observations_df.columns:
            observations_df.drop('J', axis=1, inplace=True)
        if 'T' in observations_df.columns:
            observations_df.drop('T', axis=1, inplace=True)
        observations_df = pd.concat([observations_df, sampled_jt], axis=1)
        return observations_df

    def sample_jt(self, total_prob_j: list, probs_t_given_j: list, numerical_error_tolerance: float = 0.001) -> pd.DataFrame:
        """
        Sample event type and event time for each observation.

        Args:
            total_prob_j (list): A list of dataframes, one for each event type, with the total probability of event occurrance to each of the observations.
            probs_t_given_j (list): A list of dataframes, one for each event type, with the conditional probability of event occurrance at t given event type j to each of the observations.

        Returns:
            sampled_df (pd.DataFrame): A dataframe with sampled event time and event type for each observation.
        """

        total_prob_j = self._validate_prob_dfs_list(total_prob_j, numerical_error_tolerance)
        probs_t_given_j = self._validate_prob_dfs_list(probs_t_given_j, numerical_error_tolerance)

        # Add administrative censoring (no event occured until Tmax) probability as J=0
        temp_sums = pd.concat([1 - sum(total_prob_j), *total_prob_j], axis=1, keys=[0, *self.events])
        if (((temp_sums < (0 - numerical_error_tolerance)).any().any()) or \
                ((temp_sums > (1 + numerical_error_tolerance)).any().any())):
            raise ValueError("The chosen sampling parameters result in invalid probabilities")
        # Only fixes numerical errors smaller than the tolerance size
        temp_sums.clip(0, 1, inplace=True)

        # Efficient way to sample j for each observation with different event probabilities
        sampled_df = (temp_sums.cumsum(1) > np.random.rand(temp_sums.shape[0])[:, None]).idxmax(axis=1).to_frame('J')

        temp_ts = []
        for j in self.events:
            # Get the index of the observations with J_i = j
            rel_j = sampled_df.query("J==@j").index

            # Get probs dataframe from the dfs list
            prob_df = probs_t_given_j[j - 1]  # the prob j to sample from

            # Sample time of occurrence given J_i = j
            temp_ts.append((prob_df.loc[rel_j].cumsum(1) >= np.random.rand(rel_j.shape[0])[:, None]).idxmax(axis=1))

        # Add Tmax+1 for observations with J_i = 0
        temp_ts.append(pd.Series(self.d_times + 1, index=sampled_df.query('J==0').index))
        sampled_df["T"] = pd.concat(temp_ts).sort_index()
        return sampled_df

    def sample_independent_lof_censoring(self, observations_df: pd.DataFrame,
                                         prob_lof_at_t: np.array, seed: Union[int, None] = None) -> pd.DataFrame:
        """
        Samples loss of follow-up censoring time from probabilities independent of covariates.

        Args:
            observations_df (pd.DataFrame): Dataframe with observations covariates.
            prob_lof_at_t (np.array): Array of probabilities for sampling each of the possible times.
            seed (int): pseudo random seed number for numpy.random.seed()

        Returns:
            observations_df (pd.DataFrame): Upadted dataframe including sampled censoring time.
        """
        np.random.seed(seed)
        administrative_censoring_prob = (1 - sum(prob_lof_at_t))
        assert (administrative_censoring_prob >= 0), "Check the sum of prob_lof_at_t argument."
        assert (administrative_censoring_prob <= 1), "Check the sum of prob_lof_at_t argument."

        prob_lof_at_t = np.append(prob_lof_at_t, administrative_censoring_prob)
        sampled_df = pd.DataFrame(np.random.choice(a=self.times, size=len(observations_df), p=prob_lof_at_t),
                                  index=observations_df.index, columns=['C'])
        # No follow-up censoring, C=d+2 such that T wins when building X column:
        #sampled_df.loc[sampled_df['C'] == self.times[-1], 'C'] = self.d_times + 2
        if 'C' in observations_df.columns:
            observations_df.drop('C', axis=1, inplace=True)
        observations_df = pd.concat([observations_df, sampled_df], axis=1)
        return observations_df

    def sample_hazard_lof_censoring(self, observations_df: pd.DataFrame, censoring_hazard_coefs: dict,
                                    seed: Union[int, None] = None,
                                    covariates: Union[list, None] = None) -> pd.DataFrame:
        """
        Samples loss of follow-up censoring time from hazard coefficients.

        Args:
            observations_df (pd.DataFrame): Dataframe with observations covariates.
            censoring_hazard_coefs (dict): time coefficients and covariates coefficients for the censoring hazard.
            seed (int): pseudo random seed number for numpy.random.seed()
            covariates (list): list of covariates names, must be a subset of observations_df.columns

        Returns:
            observations_df (pd.DataFrame): Upadted dataframe including sampled censoring time.
        """
        if covariates is None:
            covariates = [c for c in observations_df.columns if c not in ['X', 'T', 'C', 'J']]
        cov_df = observations_df[covariates]
        tmp_ets = EventTimesSampler(d_times=self.d_times, j_event_types=1)
        sampled_df = tmp_ets.sample_event_times(cov_df, censoring_hazard_coefs, seed=seed, covariates=covariates,
                                                events=[0])

        # No follow-up censoring, C=d+2 such that T wins when building X column:
        #sampled_df.loc[sampled_df['J'] == 0, 'T'] = self.d_times + 2
        sampled_df = sampled_df[['T']]
        sampled_df.columns = ['C']
        if 'C' in observations_df.columns:
            observations_df.drop('C', axis=1, inplace=True)
        observations_df = pd.concat([observations_df, sampled_df], axis=1)
        return observations_df

    def update_event_or_lof(self, observations_df: pd.DataFrame) -> pd.DataFrame:
        """
        Updates time column 'X' to be the minimum between event time column 'T' and censoring time column 'C'.
        Event type 'J' will be changed to 0 for observation with 'C' < 'T'.

        Args:
            observations_df (pd.DataFrame): Dataframe with observations after sampling event times 'T' and censoring time 'C'.

        Returns:
            observations_df (pd.DataFrame): Dataframe with updated time column 'X' and event type column 'J'
        """
        assert ('T' in observations_df.columns), "Trying to update event or censoring before sampling event times"
        assert ('C' in observations_df.columns), "Trying to update event or censoring before sampling censoring time"
        observations_df['X'] = observations_df[['T', 'C']].min(axis=1)
        observations_df.loc[observations_df.loc[(observations_df['C'] < observations_df['T'])].index, 'J'] = 0
        return observations_df



import pandas as pd
import numpy as np

DEFAULT_N_PATIENTS = 10000

min_year = 2000
max_year = 2015

bmi_coef = 0.003
age_coef = 0.002
gender_coef = -0.15
smk_coef = 0.1
max_p = 0.65

PATIENT_NO_COL = 'ID'
AGE_COL = 'Age'
GENDER_COL = 'Sex'
ADMISSION_YEAR_COL = 'Admyear'
FIRST_ADMISSION_COL = 'Firstadm'
ADMISSION_SERIAL_COL = 'Admserial'
WEIGHT_COL = 'Weight'
HEIGHT_COL = 'Height'
BMI_COL = 'BMI'
SMOKING_COL = 'Smoking'
HYPERTENSION_COL = 'Hypertension'
DIABETES_COL = 'Diabetes'
ART_FIB_COL = 'AF'  # Arterial Fibrillation
COPD_COL = 'COPD'  # Chronic Obstructive Pulmonary Disease
CRF_COL = 'CRF'  # Chronic Renal Failure
IN_HOSPITAL_DEATH_COL = 'In_hospital_death'
DISCHARGE_RELATIVE_COL = 'Discharge_relative_date'
DEATH_RELATIVE_COL = 'Death_relative_date_in_hosp'
DEATH_MISSING_COL = 'Death_date_in_hosp_missing'
RETURNING_PATIENT_COL = 'Returning_patient'

COEFS = pd.Series({
        AGE_COL: 0.1,
        GENDER_COL: -0.1,
        BMI_COL: 0.2,
        SMOKING_COL: 0.2,
        HYPERTENSION_COL: 0.2,
        DIABETES_COL: 0.2,
        ART_FIB_COL: 0.2,
        COPD_COL: 0.2,
        CRF_COL: 0.2,
})

ADMISSION_SERIAL_BINS = [0, 1.5, 4.5, 8.5, 21]
ADMISSION_SERIAL_LABELS = [0, 1, 2, 3]
SIMULATED_DATA_FILENAME = 'simulated_data.csv'

preconditions = [SMOKING_COL, HYPERTENSION_COL, DIABETES_COL, ART_FIB_COL, COPD_COL, CRF_COL]
font_sz = 14
title_sz = 18
AGE_BINS = list(range(0, 125, 5))
AGE_LABELS = [f'{AGE_BINS[a]}' for a in range(len(AGE_BINS)-1)]

DEFAULT_REAL_COEF_DICT = {
    "alpha": {
        1: lambda t: -1 - 0.3 * np.log(t),
        2: lambda t: -1.75 - 0.15 * np.log(t)
    },
    "beta": {
        1: -np.log([0.8, 3, 3, 2.5, 2]),
        2: -np.log([1, 3, 4, 3, 2])
    }
}



from typing import Iterable, Tuple

import numpy as np
from .simulations_data_config import *
from matplotlib import pyplot as plt
import matplotlib.gridspec as gridspec
import seaborn as sns
from lifelines import KaplanMeierFitter
from ..config import *
import os
import string
import warnings
warnings.filterwarnings('ignore')


def add_panel_text(ax, text, xplace=-0.15, fsz=17):
    ax.text(xplace, 1.1, text, transform=ax.transAxes, fontsize=fsz,
            fontweight='bold', va='top', ha='right')


def plot_first_model_coefs(models, times, train_df, n_cov=5, filename=None):
    fig, axes = plt.subplots(1, 2, figsize=(14, 6))
    ax = axes[0]
    ax.tick_params(axis='both', which='major', labelsize=15)
    ax.tick_params(axis='both', which='minor', labelsize=15)
    add_panel_text(ax=ax, text='a')
    ax.set_title(r'$\alpha_{jt}$', fontsize=26)
    ax.scatter(times, models[1].params[:len(times)].values, label='J=1 (Pred)', color='tab:blue')
    ax.plot(times, -1 -0.3*np.log(times), label='J=1 (True)', ls='--', color='tab:blue')
    ax.scatter(times, models[2].params[:len(times)].values, label='J=2 (Pred)', color='tab:green')
    ax.plot(times, -1.75 -0.15*np.log(times), label='J=2 (True)', ls='--', color='tab:green')
    ax.set_xlabel(r'Time', fontsize=18)
    ax.set_ylabel(r'$\alpha_{t}$', fontsize=18)
    ax.set_ylim([-3, 0.5])
    ax.legend(loc='upper center', fontsize=12)
    ax2 = ax.twinx()
    tmp_ajt = train_df[train_df['J'] == 1].groupby('X')['pid'].count()
    ax2.bar(tmp_ajt.index, tmp_ajt.values, label='J=1', color='tab:red', alpha=0.4, width=0.5)
    tmp_ajt = train_df[train_df['J'] == 2].groupby('X')['pid'].count()
    ax2.bar(tmp_ajt.index, tmp_ajt.values, label='J=2', color='tab:brown', alpha=0.6, align='edge', width=0.5)
    ax2.set_ylabel('Number of observed events', fontsize=16, color='red')
    ax2.tick_params(axis='y', colors='red')
    ax2.legend(loc='upper right', fontsize=12)

    ax = axes[1]
    ax.tick_params(axis='both', which='major', labelsize=15)
    ax.tick_params(axis='both', which='minor', labelsize=15)
    add_panel_text(ax=ax, text='b')
    ax.set_title(r'$\beta_{j}$', fontsize=26)
    ax.set_xlabel(r'Covariate', fontsize=18)
    ax.set_ylabel(r'$\beta}$', fontsize=18)
    ax.bar(np.arange(1, n_cov+1), models[1].params[-n_cov:], label='J=1 (Pred)', width=0.3, alpha=0.4, color='tab:blue')
    ax.scatter(-0.2+np.arange(1, n_cov+1), -np.log([0.8, 3, 3, 2.5, 2]), color='tab:blue', label='J=1 (True)',
               marker="4", s=130)
    ax.bar(np.arange(1, n_cov+1), models[2].params[-n_cov:], color='tab:green', label='J=2 (Pred)', align='edge',
           width=0.3, alpha=0.4)
    ax.scatter(0.35+np.arange(1, n_cov+1), -np.log([1, 3, 4, 3, 2]), color='tab:green', label='J=2 (True)', marker="3",
               s=130)
    ax.legend(loc='upper center', fontsize=12)
    ax.set_ylim([-1.5, 1])
    fig.tight_layout()
    if filename is not None:
        fig.savefig(os.path.join(OUTPUT_DIR, filename), dpi=300)


def plot_second_model_coefs(alpha_df, beta_models, times, n_cov=5, filename=None):
    fig, axes = plt.subplots(1, 2, figsize=(14, 6))
    ax = axes[0]
    ax.tick_params(axis='both', which='major', labelsize=15)
    ax.tick_params(axis='both', which='minor', labelsize=15)
    add_panel_text(ax=ax, text='a')
    ax.set_title(r'$\alpha_{jt}$', fontsize=26)
    tmp_ajt = alpha_df[alpha_df['J'] == 1]
    ax.scatter(tmp_ajt['X'].values, tmp_ajt['alpha_jt'].values, label='J=1 (two-step)', color='tab:blue')
    ax.plot(times, -1 -0.3*np.log(times), label='J=1 (True)', ls='--', color='tab:blue' )
    tmp_ajt = alpha_df[alpha_df['J'] == 2]
    ax.scatter(tmp_ajt['X'].values, tmp_ajt['alpha_jt'].values, label='J=2 (two-step)', color='tab:green')
    ax.plot(times, -1.75 -0.15*np.log(times), label='J=2 (True)', ls='--', color='tab:green')
    ax.set_xlabel(r'Time', fontsize=18)
    ax.set_ylabel(r'$\alpha_{t}$', fontsize=18)
    ax.legend(loc='upper center', fontsize=12)
    ax.set_ylim([-3, 0.5])
    ax2 = ax.twinx()
    tmp_ajt = alpha_df[alpha_df['J'] == 1].groupby('X')['n_jt'].sum()
    ax2.bar(tmp_ajt.index, tmp_ajt.values, label='J=1', color='tab:red', alpha=0.4, width=0.5)
    tmp_ajt = alpha_df[alpha_df['J'] == 2].groupby('X')['n_jt'].sum()
    ax2.bar(tmp_ajt.index, tmp_ajt.values, label='J=2', color='tab:brown', alpha=0.6, align='edge', width=0.5)
    ax2.legend(loc='upper right', fontsize=12)
    ax2.set_ylabel('Number of observed events', fontsize=16, color='red')
    ax2.tick_params(axis='y', colors='red')

    ax = axes[1]
    ax.tick_params(axis='both', which='major', labelsize=15)
    ax.tick_params(axis='both', which='minor', labelsize=15)
    add_panel_text(ax=ax, text='b')
    ax.set_title(r'$\beta_{j}$', fontsize=26)
    ax.bar(np.arange(1, n_cov+1), beta_models[1].params_.values, label='J=1 (two-step)', width=0.3, alpha=0.4,
           color='tab:blue')
    ax.scatter(-0.2+np.arange(1, n_cov+1), -np.log([0.8, 3, 3, 2.5, 2]), color='tab:blue', label='J=1 (True)',
               marker="4", s=130)
    ax.bar(np.arange(1, n_cov+1), beta_models[2].params_.values, color='tab:green', label='J=2 (two-step)', align='edge',
           width=0.3, alpha=0.4)
    ax.scatter(0.35+np.arange(1, n_cov+1), -np.log([1, 3, 4, 3, 2]), color='tab:green', label='J=2 (True)', marker="3",
               s=130)
    ax.legend(loc='upper center', fontsize=12)
    ax.set_xlabel('Covariate', fontsize=18)
    ax.set_ylabel(r'$\beta}$', fontsize=18)
    ax.set_ylim([-1.5, 1])
    fig.tight_layout()
    ax.legend(loc='upper center', fontsize=14)
    fig.tight_layout()
    if filename is not None:
        fig.savefig(os.path.join(OUTPUT_DIR, filename), dpi=300)


def plot_models_coefficients(alpha_dict: dict, beta_dict: dict, times: Iterable,
                             counts_df: pd.DataFrame,
                             n_cov: int = 5,
                             first_model_name: str = 'Lee et al.',
                             second_model_name: str = 'two-step',
                             filename: str = None,
                             show: bool = True) -> None:
    """
    This method takes the repetitive runs results and plotting the comparison between the methods coefs

    Args:
        alpha_dict (dict): a dict that contains for each event type (key) a dataframe of all the $\alpha_t$ (value)
        beta_dict (dict): a dict that contains for each event type(key) a dataframe of all the $\beta_t$ (value)
        times (Iterable): array like that contains all the unique times that were used
        counts_df (pandas.DataFrame): pandas dataframe which contains how many events per each time t
        n_cov (int): number of covariates (used to plot beta plot)
        first_model_name (Optional[str]): the name of the first model
        second_model_name (Optional[str]): the name of the second model

    Returns:
        None
    """
    fig, axes = plt.subplots(1, 2, figsize=(14, 6))
    ax = axes[0]
    ax.tick_params(axis='both', which='major', labelsize=15)
    ax.tick_params(axis='both', which='minor', labelsize=15)
    add_panel_text(ax=ax, text='a')
    ax.set_title(r'$\alpha_{jt}$', fontsize=26)
    tmp_ajt = alpha_dict[1]
    ax.scatter(times, tmp_ajt[f'Lee_mean'].values,
               label=f'J=1 ({first_model_name})', color='tab:blue', marker='o', alpha=0.4, s=40)
    ax.scatter(times, tmp_ajt[f'Ours_mean'].values,
               label=f'J=1 ({second_model_name})', color='navy', marker='*', alpha=0.7, s=40)
    ax.plot(times, tmp_ajt['real_mean'].values, label='J=1 (True)', ls='--', color='tab:blue')
    tmp_ajt = alpha_dict[2]
    ax.scatter(times, tmp_ajt[f'Lee_mean'].values,
               label=f'J=2 ({first_model_name})', color='tab:green', alpha=0.4, s=30)
    ax.scatter(times, tmp_ajt[f'Ours_mean'].values,
               label=f'J=2 ({second_model_name})', color='darkgreen', marker='*', alpha=0.7, s=30)
    ax.plot(times, tmp_ajt['real_mean'].values, label='J=2 (True)', ls='--', color='tab:green')
    ax.set_xlabel(r'Time', fontsize=18)
    ax.set_ylabel(r'$\alpha_{t}$', fontsize=18)
    ax.legend(loc='upper center', fontsize=12)
    ax.set_ylim([-3, 0.5])
    ax2 = ax.twinx()
    ax2.bar(times, counts_df.loc[1].values.squeeze(), label='J=1', color='tab:red', alpha=0.4, width=0.5)
    ax2.bar(times, counts_df.loc[2].values.squeeze(), label='J=2', color='tab:brown', alpha=0.6, align='edge',
            width=0.5)
    ax2.legend(loc='upper right', fontsize=12)
    ax2.set_ylabel('Number of observed events', fontsize=16, color='red')
    ax2.tick_params(axis='y', colors='red')

    ax = axes[1]
    ax.tick_params(axis='both', which='major', labelsize=15)
    ax.tick_params(axis='both', which='minor', labelsize=15)
    ax.set_title(r'$\beta_{j}$', fontsize=26)
    add_panel_text(ax=ax, text='b')
    beta_j = beta_dict[1]
    ax.bar(np.arange(1, n_cov + 1), beta_j['real_mean'].values, label='J=1 (True)', width=0.3, alpha=0.4,
           color='tab:blue')
    ax.scatter(-0.2 + np.arange(1, n_cov + 1), beta_j[f'Lee_mean'].values,
               color='tab:blue', label=f'J=1 ({first_model_name})', marker="4", s=130)
    ax.scatter(-0.2 + np.arange(1, n_cov + 1), beta_j[f'Ours_mean'].values,
               color='navy', label=f'J=1 ({second_model_name})', marker=">", s=130, alpha=0.4)

    beta_j = beta_dict[2]
    ax.bar(np.arange(1, n_cov + 1), beta_j['real_mean'].values, color='tab:green', label='J=2 (True)', align='edge',
           width=0.3, alpha=0.4)
    ax.scatter(0.35 + np.arange(1, n_cov + 1), beta_j[f'Lee_mean'].values,
               color='tab:green', label=f'J=2 ({first_model_name})', marker="3",
               s=130)
    ax.scatter(0.35 + np.arange(1, n_cov + 1), beta_j[f'Ours_mean'].values,
               color='darkgreen', label=f'J=2 ({second_model_name})', marker="<",
               s=130, alpha=0.4)
    ax.set_xlabel('Covariate', fontsize=18)
    ax.set_ylabel(r'$\beta}$', fontsize=18)
    ax.legend(loc='upper center', fontsize=12)
    ax.set_ylim([-1.5, 1])
    fig.tight_layout()
    if filename is not None:
        fig.savefig(filename, dpi=300)
    if show:
        plt.show()


def plot_LOS_simulation_figure1(data_df):
    text_sz = 16

    fig, axes = plt.subplots(2, 2, figsize=(14, 8))

    ax = axes[0, 0]
    add_panel_text(ax=ax, text='a')
    data_df.groupby([ADMISSION_YEAR_COL, DEATH_MISSING_COL]).size().unstack().plot(kind='bar', stacked=True, ax=ax)
    ax.set_ylabel('Number of patients', fontsize=font_sz)
    ax.set_xlabel('Year of admission', fontsize=font_sz)
    ax.legend(labels=['Yes', 'No'], title="In hospital death", fontsize='small', fancybox=True)
    ax.set_ylim([0, 1000])

    ax = axes[0, 1]
    add_panel_text(ax=ax, text='b')
    tmp = data_df[[AGE_COL, GENDER_COL]]
    tmp[AGE_COL] = pd.cut(tmp[AGE_COL], bins=AGE_BINS, labels=AGE_LABELS)
    tmp.groupby([AGE_COL, GENDER_COL]).size().unstack().plot(kind='bar', ax=ax)
    ax.set_xlabel('Age [years]', fontsize=font_sz)
    ax.set_ylabel('Number of patients', fontsize=font_sz)
    ax.legend(labels=['Male', 'Female'], title="Sex")
    ax.set_xticklabels(AGE_LABELS, rotation=90)

    ax = axes[1, 0]
    add_panel_text(ax=ax, text='c')
    ser = data_df.sort_values(by=[PATIENT_NO_COL, ADMISSION_SERIAL_COL]).drop_duplicates(
        subset=[PATIENT_NO_COL], keep='last')[ADMISSION_SERIAL_COL]
    sns.distplot(ser, kde=False, ax=ax)
    ax.set_yscale('log')
    ax.set_ylabel('Number of patients', fontsize=font_sz)
    ax.set_xlabel('Number of admissions', fontsize=font_sz)
    ax.set_xticks(list(range(1, int(data_df[ADMISSION_SERIAL_COL].max()+1))))
    ax.set_xticklabels(list(range(1, int(data_df[ADMISSION_SERIAL_COL].max()+1))))

    ax.axvline(1.5, ls='--', color='r')
    ax.axvline(4.5, ls='--', color='r')
    ax.axvline(8.5, ls='--', color='r')
    ax.text(x=0.5, y=1100, s='0', fontsize=text_sz, color='r')
    ax.text(x=2.5, y=1100, s='1', fontsize=text_sz, color='r')
    ax.text(x=6.5, y=1100, s='2', fontsize=text_sz, color='r')
    ax.text(x=10, y=1100, s='3', fontsize=text_sz, color='r')
    ax.text(x=9, y=200, s='Returning patient\ngroup', fontsize=text_sz, color='r')

    ax.grid('both', 'both')

    ax = axes[1, 1]
    add_panel_text(ax=ax, text='d')
    kmf = KaplanMeierFitter(label='In-hospital death censoring')
    T = data_df[DISCHARGE_RELATIVE_COL].values
    E = data_df[DEATH_RELATIVE_COL].isnull().astype(int).values
    kmf.fit(durations=T, event_observed=E)
    kmf.plot_survival_function(ax=ax)

    kmf = KaplanMeierFitter(label='No censoring')
    T = data_df[DISCHARGE_RELATIVE_COL].values
    E = np.ones_like(T)
    kmf.fit(durations=T, event_observed=E)
    kmf.plot_survival_function(ax=ax)

    ax.set_ylabel('Population', fontsize=font_sz)
    ax.set_xlabel('LOS (Days)', fontsize=font_sz)
    ax.set_xlim([0, 30])
    ax.grid()

    fig.tight_layout()

def plot_LOS_simulation_figure2(data_df):
    tmp = data_df.copy()
    tmp['binned_age'] = pd.cut(tmp[AGE_COL], bins=AGE_BINS, labels=AGE_LABELS)
    tmp['death_at_hosp_ind'] = tmp[DEATH_RELATIVE_COL].notnull().astype(int)

    max_time = 30
    total_ad = len(data_df)
    ihd = (tmp.groupby(DEATH_RELATIVE_COL).size() / total_ad).reindex(range(0, max_time + 1)).fillna(0)
    ihd = ihd.loc[:max_time].cumsum()
    released = ((tmp.groupby(DISCHARGE_RELATIVE_COL).size() -
                 tmp.groupby(DEATH_RELATIVE_COL).size()) / total_ad).reindex(range(0, max_time + 1)).fillna(0)
    released = released.loc[:max_time].cumsum()
    o = np.ones_like(released)

    fig, axes = plt.subplots(2, 2, figsize=(12, 7))

    text_sz = 12
    every_nth = 5

    ax = axes[0, 0]
    add_panel_text(ax=ax, text='a')
    ax.fill_between(x=released.index, y1=(o - released.values - ihd.values), color='b', alpha=0.4)
    ax.fill_between(x=released.index, y1=o, y2=(o - released.values), color='g', alpha=0.4)
    ax.fill_between(x=released.index, y1=(o - released.values), y2=(o - released.values - ihd.values), color='r',
                    alpha=0.4)
    ax.text(x=1, y=0.2, s='Hospitalized', fontsize=text_sz)
    ax.text(x=20, y=0.8, s='Released', fontsize=text_sz)
    ax.text(x=20, y=0.3, s='Dead', fontsize=text_sz)
    ax.set_ylim([0, 1])
    ax.set_xlim([0, max_time])
    ax.set_xlabel('LOS (Days)', fontsize=font_sz)
    ax.set_ylabel('Patient status ratio', fontsize=font_sz)

    ax = axes[0, 1]
    add_panel_text(ax=ax, text='b')
    tmp[tmp['death_at_hosp_ind'] == 1].groupby(['binned_age', GENDER_COL]).size().unstack().plot(kind='bar', ax=ax)
    ax.set_xlabel('Age [years]', fontsize=font_sz)
    ax.set_ylabel('Number of patients', fontsize=font_sz)
    ax.set_xticklabels(AGE_LABELS, rotation=90)
    ax.set_title('In-hospital Death', fontsize=title_sz)

    ax = axes[1, 0]
    add_panel_text(ax=ax, text='c')
    ser = tmp[tmp['death_at_hosp_ind'] != 1].groupby(DISCHARGE_RELATIVE_COL).size()
    ser = ser.reindex(range(0, max_time + 1)).fillna(0)
    ser.plot(kind='bar', ax=ax)
    ax.set_xlim([0, max_time])
    ax.set_ylabel('Released', fontsize=font_sz)
    ax.set_xlabel('LOS (Days)', fontsize=font_sz)
    ax.grid(axis='y')

    for idl, label in enumerate(ax.xaxis.get_ticklabels()):
        if (idl % every_nth) > 0:
            label.set_visible(False)

    ax = axes[1, 1]
    add_panel_text(ax=ax, text='d')
    ser = tmp[tmp['death_at_hosp_ind'] == 1].groupby(DEATH_RELATIVE_COL).size()
    ser = ser.reindex(range(0, max_time + 1)).fillna(0)
    ser.plot(kind='bar', ax=ax)
    ax.set_xlim([0, max_time])
    ax.set_ylabel('Died', fontsize=font_sz)
    ax.set_xlabel('LOS (Days)', fontsize=font_sz)
    ax.grid(axis='y')

    for idl, label in enumerate(ax.xaxis.get_ticklabels()):
        if (idl % every_nth) > 0:
            label.set_visible(False)

    # ax = axes[2, 0]
    # add_panel_text(ax=ax, text='e')
    # ser = tmp[tmp['death_at_hosp_ind'] != 1].groupby(DISCHARGE_RELATIVE_COL).size()
    # ser = ser.reindex(range(0, max_time + 1)).fillna(0)
    # ser.plot(kind='bar', logy=True, ax=ax)
    # ax.set_xlim([0, max_time])
    # ax.set_ylabel('Released', fontsize=font_sz)
    # ax.set_xlabel('Days from Hospitalization', fontsize=font_sz)
    # ax.grid(axis='y')
    # ax.set_ylim([0, 1000])
    #
    # for idl, label in enumerate(ax.xaxis.get_ticklabels()):
    #     if (idl % every_nth) > 0:
    #         label.set_visible(False)
    #
    # ax = axes[2, 1]
    # add_panel_text(ax=ax, text='f')
    # ser = tmp[tmp['death_at_hosp_ind'] == 1].groupby(DEATH_RELATIVE_COL).size()
    # ser = ser.reindex(range(0, max_time + 1)).fillna(0)
    # ser.plot(kind='bar', logy=True, ax=ax)
    # ax.set_xlim([0, max_time])
    # ax.set_ylabel('Died', fontsize=font_sz)
    # ax.set_xlabel('Days from Hospitalization', fontsize=font_sz)
    # ax.grid(axis='y')
    # ax.set_ylim([0, 1000])
    #
    # for idl, label in enumerate(ax.xaxis.get_ticklabels()):
    #     if (idl % every_nth) > 0:
    #         label.set_visible(False)

    fig.tight_layout()

def plot_LOS_simulation_figure3(data_df):
    fig, ax = plt.subplots(1, 1, figsize=(6, 4))

    missingness_cols = [WEIGHT_COL]
    missingness_titles = [WEIGHT_COL]
    cols = missingness_cols
    tmp = data_df[cols + [ADMISSION_YEAR_COL]]
    tmp['missing'] = tmp.apply(lambda row: row.isnull().any(), axis=1)
    tmp.groupby([ADMISSION_YEAR_COL, 'missing']).size().unstack().plot(kind='bar', stacked=True, ax=ax)
    ax.set_ylabel('Number of patients', fontsize=font_sz)
    ax.set_xlabel('Year of admission', fontsize=font_sz)
    ax.set_title(missingness_titles[0], fontsize=title_sz)
    ax.set_ylim([0, 1100])
    fig.tight_layout()


def plot_LOS_simulation_desc_figure(data_df):
    text_sz = 16

    fig, axes = plt.subplots(2, 2, figsize=(14, 8))

    ax = axes[0, 0]
    ax.tick_params(axis='both', which='major', labelsize=text_sz-2)
    ax.tick_params(axis='both', which='minor', labelsize=text_sz-2)
    add_panel_text(ax=ax, text='a')
    tmp = data_df[[AGE_COL, GENDER_COL]]
    tmp[AGE_COL] = pd.cut(tmp[AGE_COL], bins=AGE_BINS, labels=AGE_LABELS)
    tmp.groupby([AGE_COL, GENDER_COL]).size().unstack().plot(kind='bar', ax=ax)
    ax.set_xlabel('Age (years)', fontsize=font_sz)
    ax.set_ylabel('Number of patients', fontsize=font_sz)
    ax.legend(labels=['Male', 'Female'], title="Sex")
    ax.set_xticklabels(AGE_LABELS, rotation=90)

    ax = axes[0, 1]
    ax.tick_params(axis='both', which='major', labelsize=text_sz-2)
    ax.tick_params(axis='both', which='minor', labelsize=text_sz-2)
    add_panel_text(ax=ax, text='b')
    tmp = data_df[[SMOKING_COL, HYPERTENSION_COL, DIABETES_COL, ART_FIB_COL, COPD_COL, CRF_COL]].sum(axis=1).to_frame()
    tmp.columns = ['pre']
    h = tmp.groupby('pre').size()
    ax.bar(h.index, h.values, width=0.4)
    ax.set_xlabel('Number of preconditions', fontsize=font_sz)
    ax.set_ylabel('Number of patients', fontsize=font_sz)

    tmp = data_df.copy()
    tmp['binned_age'] = pd.cut(tmp[AGE_COL], bins=AGE_BINS, labels=AGE_LABELS)
    tmp['death_at_hosp_ind'] = tmp[DEATH_RELATIVE_COL].notnull().astype(int)

    ax = axes[1, 0]
    ax.tick_params(axis='both', which='major', labelsize=text_sz-2)
    ax.tick_params(axis='both', which='minor', labelsize=text_sz-2)
    add_panel_text(ax=ax, text='c')
    ser = tmp[tmp['death_at_hosp_ind'] != 1].groupby(DISCHARGE_RELATIVE_COL).size()
    ser = ser.reindex(range(0, 31)).fillna(0)
    ser.plot(kind='bar', ax=ax)
    ax.set_xlim([0, 30])
    ax.set_ylabel('Released', fontsize=font_sz)
    ax.set_xlabel('LOS (Days)', fontsize=font_sz)
    ax.grid(axis='y')

    ax = axes[1, 1]
    ax.tick_params(axis='both', which='major', labelsize=text_sz-2)
    ax.tick_params(axis='both', which='minor', labelsize=text_sz-2)
    add_panel_text(ax=ax, text='d')
    ser = tmp[tmp['death_at_hosp_ind'] == 1].groupby(DEATH_RELATIVE_COL).size()
    ser = ser.reindex(range(0, 31)).fillna(0)
    ser.plot(kind='bar', ax=ax)
    ax.set_ylabel('Died', fontsize=font_sz)
    ax.set_xlabel('LOS (Days)', fontsize=font_sz)
    ax.grid(axis='y')
    ax.set_xlim([0, 30])

    fig.tight_layout()


def compare_beta_models_for_example(first_models: dict, second_models: dict,
                                    n_cov: int = 5, real_coef_dict: dict = None) -> dict:

    from pydts.utils import compare_models_coef_per_event
    models_dict = {
        "alpha": {},
        "beta": {}
    }
    assert real_coef_dict is not None, "The user should supply the coefficients of the experiment"
    for event in first_models.keys():
        for model_type in models_dict.keys():
            if model_type == "alpha":
                first_slicing = slice(-n_cov)    # for alpha, similar to [:-n_cov]
                first_model = first_models[event].params[first_slicing].copy()
                first_model.index = first_model.index.str.replace(r"\D+", "", regex=True)
                first_model = first_model.add_prefix("a")
                first_model = first_model.iloc[:-1]
                second_model = second_models[event][1][["X", "alpha_jt"]].copy()
                second_model = second_model.set_index("X")["alpha_jt"].add_prefix("a")
                real_coef = real_coef_dict[model_type][event](np.arange(1, first_model.index.shape[0] + 1))
            else:
                first_slicing = slice(-n_cov, None)  # for beta, similar to [-n_cov:]
                first_model = first_models[event].params[first_slicing].copy()
                second_model = second_models[event][0].params_ # beta
                real_coef = real_coef_dict[model_type][event]
            models_dict[model_type][event] = compare_models_coef_per_event(first_model=first_model,
                                                                           second_model=second_model,
                                                                           real_values=real_coef,
                                                                           event=event,
                                                                           first_model_label="Lee",
                                                                           second_model_label="Ours"
                                                                           )
    return models_dict


def plot_reps_coef_std(rep_dict: dict, return_summary: bool = True, filename: str = None, paper_plots: bool = False):
    alphabet_list = list(string.ascii_lowercase)
    first_key = next(iter(rep_dict))    # deal with cases where there isn't 0 in samples
    coef_types = list(rep_dict[first_key].keys())  # alpha, beta
    event_types = rep_dict[first_key][coef_types[0]].keys()
    mapping = {t: i for i, t in enumerate(coef_types)}
    fig, axes = plt.subplots(len(coef_types), len(event_types), figsize=(12, 10))
    res_dict = {coef: {event_type: None for event_type in event_types} for coef in coef_types}
    for idct, coef_type in enumerate(coef_types):
        for idet, event_type in enumerate(event_types):
            ax = axes[mapping[coef_type]][event_type - 1]
            add_panel_text(ax=ax, text=alphabet_list[idct*len(event_types)+idet])
            ax.tick_params(axis='both', which='major', labelsize=15)
            ax.tick_params(axis='both', which='minor', labelsize=15)
            df = pd.concat([dfs[coef_type][event_type] for dfs in rep_dict.values()])
            temp_df = df.groupby(df.index).agg(["mean", "std"])
            prefix = "a" if coef_type == "alpha" else "Z"
            temp_df = temp_df.loc[[f"{prefix}{idx}_{event_type}" for idx in range(1, temp_df.shape[0]+1)]]
            temp_df.columns = temp_df.columns.get_level_values(0) + "_" + temp_df.columns.get_level_values(1)
            res_dict[coef_type][event_type] = temp_df.copy()
            temp_df.plot(x="Lee_std", y="Ours_std", kind="scatter", ax=ax)
            ax.set_xlabel("Lee et al. std", fontsize=18)
            ax.set_ylabel("two-step std", fontsize=18)
            ax.plot([0, 1], [0, 1], "--", transform=ax.transAxes, alpha=0.3, color="tab:green");
            ax.grid()
            if paper_plots:
                if ((idct == 0) and (idet == 0)):
                    ax.set_xlim([0, 0.2])
                    ax.set_ylim([0, 0.2])
                elif ((idct == 0) and (idet == 1)):
                    ax.set_xlim([0, 0.25])
                    ax.set_ylim([0, 0.25])
                elif ((idct == 1) and (idet == 0)):
                    ax.set_xlim([0.025, 0.035])
                    ax.set_ylim([0.025, 0.035])
                elif ((idct == 1) and (idet == 1)):
                    ax.set_xlim([0.035, 0.05])
                    ax.set_ylim([0.035, 0.05])
            latter = "\\alpha" if coef_type == "alpha" else "\\beta"
            ax.set_title(f"${latter}{event_type}$", fontsize=18)
    fig.tight_layout()
    fig.show()
    if filename is not None:
        fig.savefig(os.path.join(OUTPUT_DIR, filename), dpi=300)

    if return_summary:
        return res_dict


def plot_jss_reps_coef_std(rep_dict: dict, return_summary: bool = True, filename: str = None, show: bool = True):
    alphabet_list = list(string.ascii_lowercase)
    first_key = next(iter(rep_dict))    # deal with cases where there isn't 0 in samples
    coef_types = list(rep_dict[first_key].keys())  # alpha, beta
    event_types = rep_dict[first_key][coef_types[0]].keys()
    fig, axes = plt.subplots(1, 2, figsize=(12, 5))
    res_dict = {coef: {event_type: None for event_type in event_types} for coef in coef_types}
    for idct, coef_type in enumerate(coef_types):
        for idet, event_type in enumerate(event_types):
            if coef_type == 'alpha':
                ax = axes[event_type - 1]
                add_panel_text(ax=ax, text=alphabet_list[idct*len(event_types)+idet])
                ax.tick_params(axis='both', which='major', labelsize=15)
                ax.tick_params(axis='both', which='minor', labelsize=15)
            df = pd.concat([dfs[coef_type][event_type] for dfs in rep_dict.values()])
            temp_df = df.groupby(df.index).agg(["mean", "std"])
            prefix = "a" if coef_type == "alpha" else "Z"
            temp_df = temp_df.loc[[f"{prefix}{idx}_{event_type}" for idx in range(1, temp_df.shape[0]+1)]]
            temp_df.columns = temp_df.columns.get_level_values(0) + "_" + temp_df.columns.get_level_values(1)
            res_dict[coef_type][event_type] = temp_df.copy()
            if coef_type == 'alpha':
                temp_df.plot(x="Lee_std", y="Ours_std", kind="scatter", ax=ax)
                ax.set_xlabel("Lee et al. std", fontsize=18)
                ax.set_ylabel("two-step std", fontsize=18)
                ax.plot([0, 1], [0, 1], "--", transform=ax.transAxes, alpha=0.3, color="tab:green");
                ax.grid()
                if ((idct == 0) and (idet == 0)):
                    ax.set_xlim([0, 0.2])
                    ax.set_ylim([0, 0.2])
                elif ((idct == 0) and (idet == 1)):
                    ax.set_xlim([0, 0.25])
                    ax.set_ylim([0, 0.25])
                latter = "\\alpha" if coef_type == "alpha" else "\\beta"
                ax.set_title(f"${latter}{event_type}$", fontsize=18)
    fig.tight_layout()
    if show:
        plt.show()
    if filename is not None:
        fig.savefig(filename, dpi=300)
    if return_summary:
        return res_dict



def plot_times(times_dict: dict,
               filename: str = None,
               ax = None, color='tab:blue',
               show: bool = True) -> None:
    if ax is None:

        ax = pd.DataFrame.from_dict(times_dict).boxplot(figsize=(8, 6), boxprops={"lw": 1.5, "color": "tab:blue"},
                                                        medianprops={"lw": 2, "color": "tab:green"},
                                                        flierprops={'markeredgecolor': "tab:blue"})
    else:
        pd.DataFrame.from_dict(times_dict).boxplot(boxprops={"lw": 1.5, "color": color},
                                                   flierprops={'markeredgecolor': color},
                                                   medianprops={"lw": 2, "color": color}, ax=ax)

    # ax.set_ylim(0, 15)
    ax.tick_params(axis='both', which='major', labelsize=15)
    ax.tick_params(axis='both', which='minor', labelsize=15)
    ax.set_ylabel("Fitting Time [seconds]", fontdict={"size": 18}) # "weight": 'bold',
    ax.set_xlabel("Model type", fontdict={"size": 18}) # "weight": 'bold',

    #ax.tick_params(labelsize=14, grid_lw=0.5, grid_alpha=0.6)
    plt.tight_layout()
    if show:
        plt.show()
    if filename is not None:
        ax.figure.savefig(os.path.join(OUTPUT_DIR, filename), dpi=300)
    return ax


def plot_cif_plots(pred_df: pd.DataFrame, event: str, return_ax: bool = False, ax: plt.Axes = None,
                   pad: float = 0.15, scale: int = 5) -> None:
    """
    this method plot cif given pred df with cif and event
    """
    cif_cols = pred_df.columns[pred_df.columns.str.startswith("cif")]

    event_cif_cols = cif_cols[cif_cols.str.contains(f"j{event}")]

    event_x = event_cif_cols.str.extract(r"(t\d+)")[0].str.extract((r"(\d+)")).apply(pd.to_numeric).values.flatten()
    if ax is None:
        ax = pred_df[event_cif_cols].T.plot(figsize=(10, 10))
    else:
        pred_df[event_cif_cols].T.plot(ax=ax)
    ax.tick_params(axis='both', which='major', labelsize=15)
    ax.tick_params(axis='both', which='minor', labelsize=15)
    ax.set_xticks(event_x)
    ax.set_xticklabels(event_x)
    y_min, y_max = get_y_perc_limits(pred_df, event_cif_cols, pad=pad, scale=scale)
    ax.set_ylim(bottom=y_min, top=y_max)
    ax.set_xlabel("t", fontdict={'size': 18})  # , "weight": "bold"
    ax.set_ylabel(f"CIF of $J_{event}$", fontdict={'size': 18})  # , "weight": "bold"
    ax.grid()
    if return_ax:
        return ax
    else:
        plt.tight_layout()
        plt.show()


def scale_perc_limits(num: float, scale: int, up: bool = False):
    func = np.ceil if up else np.floor
    return (func((num * 100) / scale) * scale) / 100


def get_y_perc_limits(df: pd.DataFrame, cols: Iterable, pad: float = 0.15, scale: int = 5) -> Tuple[float, float]:
    y_min = max(df[cols].min().min() - pad, 0)
    y_min = scale_perc_limits(y_min, scale=scale, up=False)

    y_max = min(df[cols].max().max() + pad, 1)
    y_max = scale_perc_limits(y_max, scale=scale, up=True)
    return y_min, y_max


def plot_events_occurrence(patients_df: pd.DataFrame, ax: plt.Axes = None, event_type_col: str = 'J',
                           pid_col: str = 'pid', event_time_col: str = 'X', fname: str = None,
                           show: bool = True):
    if ax is None:
        tight = True
        fig, ax = plt.subplots(1, 1, figsize=(10, 4))
    else:
        tight = False
    patients_df.groupby([event_type_col, event_time_col])[pid_col].count().unstack('J').fillna(0).plot(ax=ax,
                                                                                                       kind='bar',
                                                                                                       width=0.8)
    ax.tick_params(axis='both', which='major', labelsize=15)
    ax.tick_params(axis='both', which='minor', labelsize=15)
    ax.set_xlabel(f"Time", fontdict={'size': 18})
    ax.set_ylabel(f"Number of Observations", fontdict={'size': 18})
    if tight:
        fig.tight_layout()
    if fname is not None:
        fig.savefig(fname, dpi=300)
    if show:
        plt.show()
    return ax


def plot_example_pred_output(pred_df, fname: str = None, show: bool = True):
    gs = gridspec.GridSpec(4, 4)
    ax1 = plt.subplot(gs[0, 0:2])
    ax2 = plt.subplot(gs[0, 2:])
    ax3 = plt.subplot(gs[1, 0:2])
    ax4 = plt.subplot(gs[1, 2:])
    ax5 = plt.subplot(gs[2, 0:2])
    ax6 = plt.subplot(gs[2, 2:])
    ax7 = plt.subplot(gs[3, 1:3])
    fig = plt.gcf()
    fig.set_size_inches(14, 14)
    ax_lst = [ax1, ax2, ax3, ax4, ax5, ax6, ax7]

    titles = ['Hazard', 'Probability', 'CIF', 'Overall Survival']
    letters = ['a', 'b', 'c', 'd', 'e', 'f', 'g']

    for idp, pref in enumerate(['hazard', 'prob', 'cif', 'overall_survival']):
        for ide, event in enumerate(['j1', 'j2']):
            if idp*2+ide > 6:
                break
            ax = ax_lst[idp*2+ide]
            add_panel_text(ax, letters[idp*2+ide])
            for patient in pred_df.columns:
                times = [m for m in range(1, 30)]
                if pref == 'overall_survival':
                    index_val = [f'{pref}_t{m}' for m in times]
                elif pref == 'hazard':
                    index_val = [f'{pref}_{event}_t{m}' for m in times]
                else:
                    index_val = [f'{pref}_{event}_at_t{m}' for m in times]
                tmp = pred_df.loc[index_val, patient]
                if pref in ['cif', 'overall_survival']:
                    ax.step(tmp.index, tmp.values, label=patient)
                else:
                    ax.plot(tmp.index, tmp.values, label=patient)
            ax.set_xticks(range(len(times)))
            ax.set_xticklabels(times, rotation=90)
            ax.set_xlabel('Time', fontsize=15)
            if idp*2+ide < 6:
                ax.set_ylabel(f'{titles[idp]}   {event}', fontsize=15)
            else:
                ax.set_ylabel(f'{titles[idp]}', fontsize=15)
                
            ax.tick_params(axis='both', which='major', labelsize=14)
            ax.tick_params(axis='both', which='minor', labelsize=14)
            ax.legend()
        
    fig.tight_layout()
    if fname is not None:
        fig.savefig(fname, dpi=300)
    if show:
        plt.show()


def plot_example_estimated_params(fitter, fname: str = None):
    fig, axes = plt.subplots(1,2, figsize=(14,7))
    ax = axes[0]
    fitter.plot_all_events_alpha(ax=ax, show=False)
    ax.grid()
    ax.legend(fontsize=16, loc='center right')
    add_panel_text(ax, 'a')
    ax = axes[1]
    fitter.plot_all_events_beta(ax=ax, show=False, xlabel='Value')
    ax.legend(fontsize=16, loc='center right')
    add_panel_text(ax, 'b')
    fig.tight_layout()
    if fname is not None:
        fig.savefig(fname, dpi=300)


def plot_sampled_covariates_figure(observations_df, fname, show=True):
    fontsize = 17
    fig, axes = plt.subplots(1, 3, figsize=(16, 4))
    ax = axes[0]
    ax.tick_params(axis='both', which='major', labelsize=15)
    ax.tick_params(axis='both', which='minor', labelsize=15)
    add_panel_text(ax=ax, text='a')
    ax.bar(observations_df['Z1'].value_counts().index, observations_df['Z1'].value_counts().values, width=0.4)
    ax.set_xlabel('Z1', fontsize=fontsize)
    ax.set_ylabel('Number of observations', fontsize=fontsize)
    ax = axes[1]
    ax.tick_params(axis='both', which='major', labelsize=15)
    ax.tick_params(axis='both', which='minor', labelsize=15)
    add_panel_text(ax=ax, text='b')
    ax.hist(observations_df[observations_df['Z1'] == 0]['Z2'], bins=np.arange(30, 130, step=2), color='tab:orange',
            label=r'$Z_1=0$', alpha=0.4)
    ax.hist(observations_df[observations_df['Z1'] == 1]['Z2'], bins=np.arange(30, 130, step=2), color='tab:green',
            label=r'$Z_1=1$', alpha=0.4)
    ax.legend(fontsize=16)
    ax.set_xlabel('Z2', fontsize=fontsize)
    ax.set_ylabel('Number of observations', fontsize=fontsize)
    ax = axes[2]
    ax.tick_params(axis='both', which='major', labelsize=15)
    ax.tick_params(axis='both', which='minor', labelsize=15)
    add_panel_text(ax=ax, text='c')
    ax.bar(observations_df['Z3'].value_counts().index, observations_df['Z3'].value_counts().values, width=0.4)
    ax.set_xlabel('Z3', fontsize=fontsize)
    ax.set_ylabel('Number of observations', fontsize=fontsize)
    fig.tight_layout()
    fig.savefig(fname, dpi=300)
    if show:
        plt.show()



import pandas as pd
import numpy as np
from itertools import product
from .fitters import TwoStagesFitter, TwoStagesFitterExact
import warnings
from copy import deepcopy
pd.set_option("display.max_rows", 500)
warnings.filterwarnings('ignore')
from time import time
from typing import Union
slicer = pd.IndexSlice
from .evaluation import events_integrated_brier_score, global_brier_score, events_integrated_auc, global_auc


class BasePenaltyGridSearch(object):

    """ This class implements the penalty parameter grid search. """

    def __init__(self):
        self.l1_ratio = None
        self.penalizers = []
        self.seed = None
        self.meta_models = {}
        self.train_df = None
        self.test_df = None
        self.global_auc = {}
        self.integrated_auc = {}
        self.global_bs = {}
        self.integrated_bs = {}
        self.TwoStagesFitter_type = 'CoxPHFitter'

    def evaluate(self,
                 train_df: pd.DataFrame,
                 test_df: pd.DataFrame,
                 l1_ratio: float,
                 penalizers: list,
                 metrics: Union[list, str] = ['IBS', 'GBS', 'IAUC', 'GAUC'],
                 seed: Union[None, int] = None,
                 event_type_col: str = 'J',
                 duration_col: str = 'X',
                 pid_col: str = 'pid',
                 twostages_fit_kwargs: dict = {}) -> tuple:

        """
        This function implements model estimation using train_df and evaluation of the metrics using test_df to all the possible combinations of penalizers.

        Args:
            train_df (pd.DataFrame): training data for fitting the model.
            test_df (pd.DataFrame): testing data for evaluating the estimated model's performance.
            l1_ratio (float): regularization ratio for the CoxPHFitter (see lifelines.fitters.coxph_fitter.CoxPHFitter documentation).
            penalizers (list): penalizer options for each event (see lifelines.fitters.coxph_fitter.CoxPHFitter documentation).
            metrics (str, list): Evaluation metrics.
            seed (int): pseudo random seed number for numpy.random.seed()
            event_type_col (str): The event type column name (must be a column in df), Right-censored sample (i) is indicated by event value 0, df.loc[i, event_type_col] = 0.
            duration_col (str): Last follow up time column name (must be a column in df).
            pid_col (str): Sample ID column name (must be a column in df).
            twostages_fit_kwargs (dict): keyword arguments to pass to the TwoStagesFitter.

        Returns:
            output (Tuple): Penalizers with best performance in terms of Global-AUC, if 'GAUC' is in metrics.

        """

        self.l1_ratio = l1_ratio
        self.penalizers = penalizers
        self.seed = seed
        np.random.seed(seed)

        for idp, penalizer in enumerate(penalizers):

            fit_beta_kwargs = self._get_model_fit_kwargs(penalizer, l1_ratio)

            if self.TwoStagesFitter_type == 'Exact':
                self.meta_models[penalizer] = TwoStagesFitterExact()
            else:
                self.meta_models[penalizer] = TwoStagesFitter()
            print(f"Started estimating the coefficients for penalizer {penalizer} ({idp+1}/{len(penalizers)})")
            start = time()
            self.meta_models[penalizer].fit(df=train_df, fit_beta_kwargs=fit_beta_kwargs,
                                            pid_col=pid_col, event_type_col=event_type_col, duration_col=duration_col,
                                            **twostages_fit_kwargs)
            end = time()
            print(f"Finished estimating the coefficients for penalizer {penalizer} ({idp+1}/{len(penalizers)}), {int(end - start)} seconds")

        events = [j for j in sorted(train_df[event_type_col].unique()) if j != 0]
        grid = [penalizers for e in events]
        penalizers_combinations = list(product(*grid))

        for idc, combination in enumerate(penalizers_combinations):
            mixed_two_stages = self.get_mixed_two_stages_fitter(combination)

            pred_df = mixed_two_stages.predict_prob_events(test_df)

            for metric in metrics:
                if metric == 'IAUC':
                    self.integrated_auc[combination] = events_integrated_auc(pred_df, event_type_col=event_type_col,
                                                                                      duration_col=duration_col)
                elif metric == 'GAUC':
                    self.global_auc[combination] = global_auc(pred_df, event_type_col=event_type_col,
                                                                       duration_col=duration_col)
                elif metric == 'IBS':
                    self.integrated_bs[combination] = events_integrated_brier_score(pred_df,
                                                                                    event_type_col=event_type_col,
                                                                                    duration_col=duration_col)
                elif metric == 'GBS':
                    self.global_bs[combination] = global_brier_score(pred_df, event_type_col=event_type_col,
                                                                              duration_col=duration_col)

        output = self.convert_results_dict_to_df(self.global_auc).idxmax().values[0] if 'GAUC' in metrics else []
        return output

    def convert_results_dict_to_df(self, results_dict):
        """
        This function converts a results dictionary to a pd.DataFrame format.

        Args:
            results_dict: one of the class attributes: global_auc, integrated_auc, global_bs, integrated_bs.

        Returns:
            df (pd.DataFrame): Results in a pd.DataFrame format.
        """
        df = pd.DataFrame(results_dict.values(), index=pd.MultiIndex.from_tuples(results_dict.keys()))
        return df

    def get_mixed_two_stages_fitter(self, penalizers_combination: list) -> TwoStagesFitter:
        """
        This function creates a mixed TwoStagesFitter from the estimated meta models for a specific penalizers combination.

        Args:
            penalizers_combination (list): List with length equals to the number of competing events. The penalizers value to each of the events. Each of the values must be one of the values that was previously passed to the evaluate() method.

        Returns:
            mixed_two_stages (pydts.fitters.TwoStagesFitter): TwoStagesFitter for the required penalty combination.
        """
        _validate_estimated_value = [p for p in penalizers_combination if p not in list(self.meta_models.keys())]
        assert len(_validate_estimated_value) == 0, \
               f"Values {_validate_estimated_value} were note estimated. All the penalizers in penalizers_combination must be estimated using evaluate() before a mixed combination can be generated."

        events = self.meta_models[penalizers_combination[0]].events
        event_type_col = self.meta_models[penalizers_combination[0]].event_type_col
        if self.TwoStagesFitter_type == 'Exact':
            mixed_two_stages = TwoStagesFitterExact()
        else:
            mixed_two_stages = TwoStagesFitter()

        for ide, event in enumerate(sorted(events)):
            if ide == 0:
                mixed_two_stages.covariates = self.meta_models[penalizers_combination[ide]].covariates
                mixed_two_stages.duration_col = self.meta_models[penalizers_combination[ide]].duration_col
                mixed_two_stages.event_type_col = self.meta_models[penalizers_combination[ide]].event_type_col
                mixed_two_stages.events = self.meta_models[penalizers_combination[ide]].events
                mixed_two_stages.pid_col = self.meta_models[penalizers_combination[ide]].pid_col
                mixed_two_stages.times = self.meta_models[penalizers_combination[ide]].times

            mixed_two_stages.beta_models[event] = self.meta_models[penalizers_combination[ide]].beta_models[event]
            mixed_two_stages.event_models[event] = []
            mixed_two_stages.event_models[event].append(self.meta_models[penalizers_combination[ide]].beta_models[event])

            event_alpha = self.meta_models[penalizers_combination[ide]].alpha_df.copy()
            event_alpha = event_alpha[event_alpha[event_type_col] == event]
            mixed_two_stages.alpha_df = pd.concat([mixed_two_stages.alpha_df, event_alpha])
            mixed_two_stages.event_models[event].append(event_alpha)

        return mixed_two_stages

    def _get_model_fit_kwargs(self, penalizer, l1_ratio):
        if self.TwoStagesFitter_type == 'Exact':
            fit_beta_kwargs = {
                'model_fit_kwargs': {
                    'alpha': penalizer,
                    'L1_wt': l1_ratio
                }
            }
        else:
            fit_beta_kwargs = {
                'model_kwargs': {
                    'penalizer': penalizer,
                    'l1_ratio': l1_ratio
                },
            }
        return fit_beta_kwargs


class PenaltyGridSearch(BasePenaltyGridSearch):

    def __init__(self):
        super().__init__()
        self.TwoStagesFitter_type = 'CoxPHFitter'


class PenaltyGridSearchExact(BasePenaltyGridSearch):

    def __init__(self):
        super().__init__()
        self.TwoStagesFitter_type = 'Exact'



import pandas as pd
import numpy as np
import warnings
pd.set_option("display.max_rows", 500)
warnings.filterwarnings('ignore')
slicer = pd.IndexSlice
from lifelines.fitters.kaplan_meier_fitter import KaplanMeierFitter
from typing import Optional, List, Union


# Weights
def event_specific_weights(pred_df: pd.DataFrame,
                           event: int,
                           event_type_col: str = 'J',
                           duration_col: str = 'X') -> pd.Series:
    """
    This function implements the calculation of the event specific time-weights.

    Args:
        pred_df (pd.DataFrame): Data to calculate prediction error. Must contain the observed duration and event-type, and the probability of event at time t prediction results for the event. See TwoStagesFitter.predict_prob_events()
        event (int): Event-type to calculate the weights for.
        duration_col (str): Last follow up time column name (must be a column in pred_df).
        event_type_col (str): The event type column name (must be a column in df), Right-censored sample (i) is indicated by event value 0, df.loc[i, event_type_col] = 0.

    Returns:
        result (pd.Series): event specific weights.
    """

    event_df = pred_df[pred_df[event_type_col] == event]
    if len(event_df) == 0:
        print(f'Could not calculate weights for event {event} - no test events')
        return np.nan
    weights = event_df.groupby(duration_col).size() / event_df.groupby(duration_col).size().sum()
    times = sorted(pred_df[duration_col].unique())[:-1]
    return weights.reindex(times).fillna(0)


# BRIER SCORE
def event_specific_brier_score_at_t(pred_df: pd.DataFrame,
                                    event: int,
                                    t: int,
                                    event_type_col: str = 'J',
                                    duration_col: str = 'X') -> float:
    """
    This function implements the calculation of the event specific Brier Score at time t.

    Args:
        pred_df (pd.DataFrame): Data to calculate Brier Score for. Must contain the observed duration and event-type, and the probability of event at time t prediction results for all events. See TwoStagesFitter.predict_prob_events()
        event (int): Event-type to calculate the Brier Score for.
        t (int): time to calculate the Brier Score for.
        duration_col (str): Last follow up time column name (must be a column in pred_df).
        event_type_col (str): The event type column name (must be a column in df), Right-censored sample (i) is indicated by event value 0, df.loc[i, event_type_col] = 0.

    Returns:
        result (pd.Series): event specific Brier Score at time t.
    """

    pi_ij = pred_df.loc[:, f'prob_j{event}_at_t{t}']
    D_ij = ((pred_df.loc[:, event_type_col] == event) & (pred_df.loc[:, duration_col] == t)).astype(int)
    censoring_kmf = KaplanMeierFitter()
    censoring_kmf.fit(durations=pred_df[duration_col], event_observed=(pred_df[event_type_col] == 0))
    in_risk_set_at_t = (pred_df.loc[:, duration_col] >= t).astype(int)
    W_ij = (in_risk_set_at_t / censoring_kmf.predict(times=t))
    BS_jt = ((W_ij*((D_ij - pi_ij)**2)).sum() / in_risk_set_at_t.sum())
    return BS_jt


def event_specific_brier_score_at_t_all(pred_df: pd.DataFrame,
                                        event: int,
                                        event_type_col: str = 'J',
                                        duration_col: str = 'X') -> pd.Series:
    """
    This function implements the calculation of the event specific Brier Score at time t for all times included in duration_col of pred_df.

    Args:
        pred_df (pd.DataFrame): Data to calculate Brier Score. Must contain the observed duration and event-type, and the probability of event at time t prediction results for the event. TwoStagesFitter.predict_prob_events()
        event (int): Event-type to calculate the Brier Score for.
        duration_col (str): Last follow up time column name (must be a column in pred_df).
        event_type_col (str): The event type column name (must be a column in df), Right-censored sample (i) is indicated by event value 0, df.loc[i, event_type_col] = 0.

    Returns:
        result (pd.Series): event specific Brier Score for all times included in duration_col of pred_df.
    """

    res = {}
    for t in sorted(pred_df[duration_col].unique())[:-1]:
        res[t] = event_specific_brier_score_at_t(pred_df=pred_df, event=event, t=t,
                                                 event_type_col=event_type_col, duration_col=duration_col)
    return pd.Series(res, name=event)


def event_specific_integrated_brier_score(pred_df: pd.DataFrame,
                                          event: int,
                                          event_type_col: str = 'J',
                                          duration_col: str = 'X',
                                          weights: Union[pd.Series, None] = None) -> float:
    """
    This function implements the calculation of the event specific integrated Brier Score.

    Args:
        pred_df (pd.DataFrame): Data to calculate Brier Score. Must contain the observed duration and event-type, and the probability of event at time t prediction results for the event. See TwoStagesFitter.predict_prob_events()
        event (int): Event-type to calculate the integrated Brier Score for.
        duration_col (str): Last follow up time column name (must be a column in pred_df).
        event_type_col (str): The event type column name (must be a column in df), Right-censored sample (i) is indicated by event value 0, df.loc[i, event_type_col] = 0.
        weights (pd.Series): Optional. Weights vector with time as index and weight as value. Length must be the number of possible event times.

    Returns:
        result (float): integrated Brier Score results.
    """

    brier_score_at_t = event_specific_brier_score_at_t_all(pred_df=pred_df, event=event,
                                                           event_type_col=event_type_col,
                                                           duration_col=duration_col)

    if brier_score_at_t.isnull().any():
        print(f'There are NaN values in BS(t) during Integrated Brier Score calculation for event {event}. Times: {brier_score_at_t[brier_score_at_t.isnull()].index}\n')

    if weights is None:
        weights = event_specific_weights(pred_df=pred_df, event=event,
                                         event_type_col=event_type_col,
                                         duration_col=duration_col)

    result = brier_score_at_t.dot(weights.sort_index())
    return result


def global_brier_score(pred_df: pd.DataFrame,
                       event_type_col: str = 'J',
                       duration_col: str = 'X') -> float:
    """
    This function implements the calculation of the global Brier Score.

    Args:
        pred_df (pd.DataFrame): Data to calculate Brier score. Must contain the observed duration and event-type, and the probability of event at time t prediction results for all events. See TwoStagesFitter.predict_prob_events()
        duration_col (str): Last follow up time column name (must be a column in pred_df).
        event_type_col (str): The event type column name (must be a column in df), Right-censored sample (i) is indicated by event value 0, df.loc[i, event_type_col] = 0.

    Returns:
        global_auc (float): global Brier Score results.
    """

    e_j_ser = pred_df[pred_df[event_type_col] != 0].groupby('J').size().sort_index()
    total_e = e_j_ser.sum()
    global_bs = 0
    for event, e_j in e_j_ser.iteritems():
        global_bs += (e_j / total_e) * event_specific_integrated_brier_score(
            pred_df=pred_df, event=event, event_type_col=event_type_col,
            duration_col=duration_col)
    return global_bs


def events_integrated_brier_score(pred_df: pd.DataFrame, event_type_col: str = 'J',
                                  duration_col: str ='X') -> dict:
    """
    This function implements the calculation of the integrated Brier Score to all events.

    Args:
        pred_df (pd.DataFrame): Data to calculate integrated Brier Score. Must contain the observed duration and event-type, and the probability of event at time t prediction results for all events. See TwoStagesFitter.predict_prob_events()
        duration_col (str): Last follow up time column name (must be a column in pred_df).
        event_type_col (str): The event type column name (must be a column in df), Right-censored sample (i) is indicated by event value 0, df.loc[i, event_type_col] = 0.

    Returns:
        integrated_brier_score (dict): integrated Brier Score results.
    """

    events = [e for e in pred_df[event_type_col].unique() if e != 0]
    integrated_brier_score = {}
    for event in sorted(events):
        integrated_brier_score[event] = event_specific_integrated_brier_score(pred_df=pred_df,
                                                                              event=event,
                                                                              event_type_col=event_type_col,
                                                                              duration_col=duration_col)
    return integrated_brier_score


def events_brier_score_at_t(pred_df: pd.DataFrame,
                            event_type_col: str = 'J',
                            duration_col: str ='X') -> pd.DataFrame:
    """
    This function implements the calculation of the events Brier score at t.

    Args:
        pred_df (pd.DataFrame): Data to calculate prediction error. Must contain the observed duration and event-type, and the probability of event at time t prediction results for all events. See TwoStagesFitter.predict_prob_events()
        duration_col (str): Last follow up time column name (must be a column in pred_df).
        event_type_col (str): The event type column name (must be a column in df), Right-censored sample (i) is indicated by event value 0, df.loc[i, event_type_col] = 0.

    Returns:
        event_brier_score_at_t_df (pd.DataFrame): events Brier score at t results.
    """

    events = [e for e in pred_df[event_type_col].unique() if e != 0]
    event_brier_score_at_t_df = pd.DataFrame()
    for event in sorted(events):
        event_brier_score_at_t_df = pd.concat([event_brier_score_at_t_df,
                                               event_specific_brier_score_at_t_all(pred_df=pred_df,
                                                                                   event=event,
                                                                                   event_type_col=event_type_col,
                                                                                   duration_col=duration_col)],
                                              axis=1)
    return event_brier_score_at_t_df.T


# AUC
def events_auc_at_t(pred_df: pd.DataFrame,
                    event_type_col: str = 'J',
                    duration_col: str ='X') -> pd.DataFrame:
    """
    This function implements the calculation of the events AUC at t.

    Args:
        pred_df (pd.DataFrame): Data to calculate AUC. Must contain the observed duration and event-type, and the probability of event at time t prediction results for all events. See TwoStagesFitter.predict_prob_events()
        duration_col (str): Last follow up time column name (must be a column in pred_df).
        event_type_col (str): The event type column name (must be a column in df), Right-censored sample (i) is indicated by event value 0, df.loc[i, event_type_col] = 0.

    Returns:
        event_auc_at_t_df (pd.DataFrame): events AUC at t results.
    """

    events = [e for e in pred_df[event_type_col].unique() if e != 0]
    event_auc_at_t_df = pd.DataFrame()
    for event in sorted(events):
        event_auc_at_t_df = pd.concat([event_auc_at_t_df,
                                       event_specific_auc_at_t_all(pred_df=pred_df,
                                                                   event=event,
                                                                   event_type_col=event_type_col,
                                                                   duration_col=duration_col)],
                                       axis=1)
    return event_auc_at_t_df.T


def event_specific_auc_at_t(pred_df: pd.DataFrame,
                            event: int,
                            t: int,
                            event_type_col: str = 'J',
                            duration_col: str = 'X') -> float:
    """
    This function implements the calculation of the event specific AUC at time t.

    Args:
        pred_df (pd.DataFrame): Data to calculate prediction error. Must contain the observed duration and event-type, and the probability of event at time t prediction results for the event. See TwoStagesFitter.predict_prob_events()
        event (int): Event-type to calculate the integrated AUC for.
        t (int): time to calculate the AUC for.
        duration_col (str): Last follow up time column name (must be a column in pred_df).
        event_type_col (str): The event type column name (must be a column in df), Right-censored sample (i) is indicated by event value 0, df.loc[i, event_type_col] = 0.

    Returns:
        result (pd.Series): event specific AUC for all times included in duration_col of pred_df.
    """

    event_observed_at_t_df = pred_df[(pred_df[event_type_col] == event) & (pred_df[duration_col] == t)]
    no_event_at_t_df = pred_df[pred_df[duration_col] >= t]
    no_event_at_t_df = no_event_at_t_df[~((no_event_at_t_df[event_type_col] == event) &
                                          (no_event_at_t_df[duration_col] == t))]
    total_t = (len(event_observed_at_t_df)*len(no_event_at_t_df))
    if total_t == 0:
        print(f'AUC could not be calculated for event {event} at time {t} - no test pairs of with and without observed event {event} at time {t}')
        return np.nan
    correct_order = 0
    for i_idx, i_row in event_observed_at_t_df.iterrows():
        pi_ij = i_row.loc[f'prob_j{event}_at_t{t}']
        pi_mj = no_event_at_t_df.loc[:, f'prob_j{event}_at_t{t}']
        correct_order += ((pi_ij > pi_mj).sum()+0.5*(pi_ij == pi_mj).sum())
    return correct_order / total_t


def event_specific_auc_at_t_all(pred_df: pd.DataFrame,
                                event: int,
                                event_type_col: str = 'J',
                                duration_col: str = 'X') -> pd.Series:
    """
    This function implements the calculation of the event specific AUC at time t for all times included in duration_col of pred_df.

    Args:
        pred_df (pd.DataFrame): Data to calculate prediction error. Must contain the observed duration and event-type, and the probability of event at time t prediction results for the event. See TwoStagesFitter.predict_prob_events()
        event (int): Event-type to calculate the AUC for.
        duration_col (str): Last follow up time column name (must be a column in pred_df).
        event_type_col (str): The event type column name (must be a column in df), Right-censored sample (i) is indicated by event value 0, df.loc[i, event_type_col] = 0.

    Returns:
        result (pd.Series): event specific AUC for all times included in duration_col of pred_df.
    """

    res = {}
    for t in sorted(pred_df[duration_col].unique())[:-1]:
        res[t] = event_specific_auc_at_t(pred_df=pred_df, event=event, t=t,
                                         event_type_col=event_type_col, duration_col=duration_col)
    return pd.Series(res, name=event)


def event_specific_integrated_auc(pred_df: pd.DataFrame,
                                  event: int,
                                  event_type_col: str = 'J',
                                  duration_col: str = 'X',
                                  weights: Union[pd.Series, None] = None) -> float:
    """
    This function implements the calculation of the event specific integrated auc.

    Args:
        pred_df (pd.DataFrame): Data to calculate prediction error. Must contain the observed duration and event-type, and the probability of event at time t prediction results for the event. See TwoStagesFitter.predict_prob_events()
        event (int): Event-type to calculate the integrated AUC for.
        duration_col (str): Last follow up time column name (must be a column in pred_df).
        event_type_col (str): The event type column name (must be a column in df), Right-censored sample (i) is indicated by event value 0, df.loc[i, event_type_col] = 0.
        weights (pd.Series): Optional. Weights vector with time as index and weight as value. Length must be the number of possible event times.

    Returns:
        result (float): integrated AUC results.
    """

    auc_at_t = event_specific_auc_at_t_all(pred_df=pred_df, event=event,
                                           event_type_col=event_type_col,
                                           duration_col=duration_col)

    if auc_at_t.isnull().any():
        print(f'There are NaN values in AUC(t) during Integrated AUC calculation for event {event}. Times: {auc_at_t[auc_at_t.isnull()].index}\n')

    if weights is None:
        if auc_at_t.isnull().any():
            print(f'Please check there are events of type {event} at each time in pred_df or provide a weights vector with weight 0 for the problematic times.')
            return np.nan
        weights = event_specific_weights(pred_df=pred_df, event=event,
                                         event_type_col=event_type_col,
                                         duration_col=duration_col)

    result = auc_at_t.dot(weights.sort_index())
    return result


def global_auc(pred_df: pd.DataFrame,
               event_type_col: str = 'J',
               duration_col: str = 'X') -> float:
    """
    This function implements the calculation of the global AUC.

    Args:
        pred_df (pd.DataFrame): Data to calculate prediction error. Must contain the observed duration and event-type, and the probability of event at time t prediction results for all events. See TwoStagesFitter.predict_prob_events()
        duration_col (str): Last follow up time column name (must be a column in pred_df).
        event_type_col (str): The event type column name (must be a column in df), Right-censored sample (i) is indicated by event value 0, df.loc[i, event_type_col] = 0.

    Returns:
        global_auc (float): global AUC results.
    """

    e_j_ser = pred_df[pred_df[event_type_col] != 0].groupby('J').size().sort_index()
    total_e = e_j_ser.sum()
    global_auc = 0
    for event, e_j in e_j_ser.iteritems():
        global_auc += (e_j / total_e) * event_specific_integrated_auc(
            pred_df=pred_df, event=event, event_type_col=event_type_col,
            duration_col=duration_col)
    return global_auc


def events_integrated_auc(pred_df: pd.DataFrame, event_type_col: str = 'J',
                          duration_col: str ='X') -> dict:
    """
    This function implements the calculation of the integrated AUC to all events.

    Args:
        pred_df (pd.DataFrame): Data to calculate prediction error. Must contain the observed duration and event-type, and the probability of event at time t prediction results for all events. See TwoStagesFitter.predict_prob_events()
        duration_col (str): Last follow up time column name (must be a column in pred_df).
        event_type_col (str): The event type column name (must be a column in df), Right-censored sample (i) is indicated by event value 0, df.loc[i, event_type_col] = 0.

    Returns:
        integrated_auc (dict): integrated AUC results.
    """

    events = [e for e in pred_df[event_type_col].unique() if e != 0]
    integrated_auc = {}
    for event in sorted(events):
        integrated_auc[event] = event_specific_integrated_auc(pred_df=pred_df,
                                                              event=event,
                                                              event_type_col=event_type_col,
                                                              duration_col=duration_col)
    return integrated_auc
